__doc__ = """
Stimjim extension.
Basic utilities for translating `spiketrains` into Stimjim pulse generator.

author: <Seung Hyun Kim skim0119@gmail.com>
"""
__all__ = ["StimjimSerial"]

from typing import List, Optional

import os
import sys
import time

import numpy as np
import serial

from miv.io.serial import ArduinoSerial
from miv.typing import SpiketrainType


class StimjimSerial(ArduinoSerial):
    """
    Module to control Stimjim using PySerial.

    All time-units are in micro.
    All volt-units are in milli.
    All ampere-units are in micro.
    """

    def __init__(
        self, port, output0_mode=1, output1_mode=3, high_v=4500, low_v=0, **kwargs
    ):
        super().__init__(port, **kwargs)
        self.output0_mode = output0_mode
        self.output1_mode = output1_mode
        self.high_v_1 = high_v
        self.low_v_1 = low_v
        self.high_v_2 = 0
        self.low_v_2 = 0

    def send_spiketrain(
        self,
        pulsetrain: int,
        spiketrain: SpiketrainType,
        t_max: int,
        total_duration: int,
        delay: float = 0.0,
        channel: int = 0,
        reverse: bool = False,
    ) -> bool:
        total_string, total_period = self._spiketrain_to_str(
            spiketrain, t_max, reverse=reverse
        )
        total_string.insert(
            0,
            self._start_str(
                pulsetrain,
                self.output0_mode,
                self.output1_mode,
                total_period,
                total_duration,
            ),
        )
        return "; ".join(total_string)

    def _start_str(self, pulsetrain, output0_mode, output1_mode, period, duration):
        return f"S{pulsetrain},{output0_mode},{output1_mode},{period},{duration}"

    def _spiketrain_to_str(
        self,
        spiketrain: SpiketrainType,
        t_max: int,
        pulse_length: int = 10_000,
        reverse: bool = False,
    ) -> List[str]:
        """
        Convert a spiketrain into a series of strings that can be sent to the Stimjim device.
        """
        spiketrain = np.insert(spiketrain, 0, 0)
        spiketrain = np.append(spiketrain, t_max)
        gaps = np.diff(spiketrain).astype(np.int_)
        if reverse:
            gaps = gaps[::-1]
        if np.any(gaps < pulse_length):
            raise ValueError(
                f"Gap between pulse {gaps} must be larger than pulse length {pulse_length}. {spiketrain}"
            )

        # String functions
        def gap_to_str(x, A1, A2):
            return f"{A1},{A2},{x:d}"

        pulse_to_str = gap_to_str(pulse_length, self.high_v_1, 0)

        total_string = [pulse_to_str, gap_to_str(gaps[0], self.low_v_1, 0)]  # First Gap
        for gap in gaps[1:]:
            total_string.append(pulse_to_str)
            total_string.append(gap_to_str(gap - pulse_length, self.low_v_1, 0))

        total_period = gaps.sum()
        return total_string, total_period



__doc__ = """

Similarity Metrics
##################

.. currentmodule:: miv.signal.similarity

.. autofunction:: domain_distance_matrix

"""

__all__ = ["domain_distance_matrix"]
from typing import Literal

import os
import sys

import numba
import numpy as np
from numba import njit

from miv.typing import SignalType

# For beyond python3.11, change to LiteralString

DOMAIN_MOD = Literal["time", "frequency", "power"]


def domain_distance_matrix(temporal_sequence: SignalType, domain: DOMAIN_MOD = "time"):
    """
    Measure similarity between signal using correlation matrix,

    Parameters
    ----------
    temporal_sequence : SignalType
        The temporal sequence (time, n_features).
    domain : Literal['time', 'frequency', 'power']
        If 'time', use time-base correlation to measure the similarity.
        If 'frequency', use FFT-base correlation to measure the similarity.
        If 'power', use power of the signal to measure the similarity.

    Returns
    -------
    matrix : np.ndarray
        2D matrix with the size (n_features, n_features)
        Each element represent the distance between the feature set.
    """
    assert (
        len(temporal_sequence.shape) > 1
    ), f"Sequence must have at least two axes. Provide sequence has a shape {temporal_sequence.shape}."

    n_time, n_features = temporal_sequence.shape
    distance_matrix = np.empty([n_features, n_features], dtype=np.float_)

    if domain == "time":
        func = nb_time_domain_similarity
    elif domain == "frequency":
        func = nb_frequency_domain_similarity
    elif domain == "power":
        func = nb_power_domain_similarity

    for i in range(n_features):
        for j in range(n_features):
            if i == j:
                distance_matrix[i, j] = 0
            else:
                similarity = func(temporal_sequence[:, i], temporal_sequence[:, j])
                distance_matrix[i, j] = similarity
                distance_matrix[j, i] = similarity

    return distance_matrix


@njit(cache=True)
def nb_time_domain_similarity(signal_a, signal_b):
    reference = np.correlate(signal_a, signal_a)
    corr = np.correlate(signal_a, signal_b)
    return np.abs(reference - corr)


# @njit(cache=True) # TODO: replace fft to other module
def nb_frequency_domain_similarity(signal_a, signal_b):
    reference = np.correlate(np.fft.fft(signal_a), np.fft.fft(signal_a))
    corr = np.correlate(np.fft.fft(signal_a), np.fft.fft(signal_b))
    return np.abs(reference - corr)


@njit(cache=True)
def nb_power_domain_similarity(signal_a, signal_b):
    reference_power = np.sum(np.square(signal_a))
    target_power = np.sum(np.square(signal_b))
    return np.abs(reference_power - target_power)



__doc__ = """
HDF5-based file format for heterogeneous numerical data.
Based on code from and inspired by

- HEPfile: https://github.com/mattbellis/hepfile
- NeuroH5: https://github.com/iraikov/neuroh5

.. automodule:: miv.io.file.read
   :members:

.. automodule:: miv.io.file.write
   :members:

"""

from miv.io.file.read import *
from miv.io.file.write import *



import logging
import math
import os
import struct
import sys
import time

import matplotlib.pyplot as plt
import numpy as np

# source: https://github.com/Intan-Technologies/load-rhs-notebook-python/tree/423aa7dd8d4287232e86fba829ac3fad12cd79ed


# Define plural function
def plural(n):  # pragma: no cover
    """Utility function to optionally pluralize words based on the value of n."""

    if n == 1:
        return ""
    else:
        return "s"


# Define read_qstring function
def read_qstring(fid):  # pragma: no cover
    """Read Qt style QString.

    The first 32-bit unsigned number indicates the length of the string (in bytes).
    If this number equals 0xFFFFFFFF, the string is null.

    Strings are stored as unicode.
    """

    (length,) = struct.unpack("<I", fid.read(4))
    if length == int("ffffffff", 16):
        return ""

    if length > (os.fstat(fid.fileno()).st_size - fid.tell() + 1):
        logging.info(length)
        raise Exception("Length too long.")

    # convert length from bytes to 16-bit Unicode words
    length = int(length / 2)

    data = []
    for i in range(0, length):
        (c,) = struct.unpack("<H", fid.read(2))
        data.append(c)

    a = "".join([chr(c) for c in data])

    return a


# Define read_header function
def read_header(fid):  # pragma: no cover
    """Reads the Intan File Format header from the given file."""

    # Check 'magic number' at beginning of file to make sure this is an Intan
    # Technologies RHD2000 data file.
    (magic_number,) = struct.unpack("<I", fid.read(4))
    if magic_number != int("0xD69127AC", 16):
        raise Exception("Unrecognized file type.")

    header = {}
    # Read version number.
    version = {}
    (version["major"], version["minor"]) = struct.unpack("<hh", fid.read(4))
    header["version"] = version

    logging.info("")
    logging.info(
        "Reading Intan Technologies RHS2000 Data File, Version {}.{}".format(
            version["major"], version["minor"]
        )
    )
    logging.info("")

    # Read information of sampling rate and amplifier frequency settings.
    (header["sample_rate"],) = struct.unpack("<f", fid.read(4))
    (
        header["dsp_enabled"],
        header["actual_dsp_cutoff_frequency"],
        header["actual_lower_bandwidth"],
        header["actual_lower_settle_bandwidth"],
        header["actual_upper_bandwidth"],
        header["desired_dsp_cutoff_frequency"],
        header["desired_lower_bandwidth"],
        header["desired_lower_settle_bandwidth"],
        header["desired_upper_bandwidth"],
    ) = struct.unpack("<hffffffff", fid.read(34))

    # This tells us if a software 50/60 Hz notch filter was enabled during
    # the data acquisition.
    (notch_filter_mode,) = struct.unpack("<h", fid.read(2))
    header["notch_filter_frequency"] = 0
    if notch_filter_mode == 1:
        header["notch_filter_frequency"] = 50
    elif notch_filter_mode == 2:
        header["notch_filter_frequency"] = 60

    (
        header["desired_impedance_test_frequency"],
        header["actual_impedance_test_frequency"],
    ) = struct.unpack("<ff", fid.read(8))
    (header["amp_settle_mode"], header["charge_recovery_mode"]) = struct.unpack(
        "<hh", fid.read(4)
    )

    frequency_parameters = {}
    frequency_parameters["amplifier_sample_rate"] = header["sample_rate"]
    frequency_parameters["board_adc_sample_rate"] = header["sample_rate"]
    frequency_parameters["board_dig_in_sample_rate"] = header["sample_rate"]
    frequency_parameters["desired_dsp_cutoff_frequency"] = header[
        "desired_dsp_cutoff_frequency"
    ]
    frequency_parameters["actual_dsp_cutoff_frequency"] = header[
        "actual_dsp_cutoff_frequency"
    ]
    frequency_parameters["dsp_enabled"] = header["dsp_enabled"]
    frequency_parameters["desired_lower_bandwidth"] = header["desired_lower_bandwidth"]
    frequency_parameters["desired_lower_settle_bandwidth"] = header[
        "desired_lower_settle_bandwidth"
    ]
    frequency_parameters["actual_lower_bandwidth"] = header["actual_lower_bandwidth"]
    frequency_parameters["actual_lower_settle_bandwidth"] = header[
        "actual_lower_settle_bandwidth"
    ]
    frequency_parameters["desired_upper_bandwidth"] = header["desired_upper_bandwidth"]
    frequency_parameters["actual_upper_bandwidth"] = header["actual_upper_bandwidth"]
    frequency_parameters["notch_filter_frequency"] = header["notch_filter_frequency"]
    frequency_parameters["desired_impedance_test_frequency"] = header[
        "desired_impedance_test_frequency"
    ]
    frequency_parameters["actual_impedance_test_frequency"] = header[
        "actual_impedance_test_frequency"
    ]

    header["frequency_parameters"] = frequency_parameters

    (
        header["stim_step_size"],
        header["recovery_current_limit"],
        header["recovery_target_voltage"],
    ) = struct.unpack("fff", fid.read(12))

    note1 = read_qstring(fid)
    note2 = read_qstring(fid)
    note3 = read_qstring(fid)
    header["notes"] = {"note1": note1, "note2": note2, "note3": note3}

    (header["dc_amplifier_data_saved"], header["eval_board_mode"]) = struct.unpack(
        "<hh", fid.read(4)
    )

    header["reference_channel"] = read_qstring(fid)

    # Create structure arrays for each type of data channel.
    header["spike_triggers"] = []
    header["amplifier_channels"] = []
    header["dc_amplifier_channels"] = []
    header["stim_channels"] = []
    header["amp_settle_channels"] = []
    header["charge_recovery_channels"] = []
    header["compliance_limit_channels"] = []
    header["board_adc_channels"] = []
    header["board_dac_channels"] = []
    header["board_dig_in_channels"] = []
    header["board_dig_out_channels"] = []

    # Read signal summary from data file header.
    (number_of_signal_groups,) = struct.unpack("<h", fid.read(2))

    for signal_group in range(1, number_of_signal_groups + 1):
        signal_group_name = read_qstring(fid)
        signal_group_prefix = read_qstring(fid)
        (
            signal_group_enabled,
            signal_group_num_channels,
            signal_group_num_amp_channels,
        ) = struct.unpack("<hhh", fid.read(6))

        if (signal_group_num_channels > 0) and (signal_group_enabled > 0):
            for signal_channel in range(0, signal_group_num_channels):
                new_channel = {
                    "port_name": signal_group_name,
                    "port_prefix": signal_group_prefix,
                    "port_number": signal_group,
                }
                new_channel["native_channel_name"] = read_qstring(fid)
                new_channel["custom_channel_name"] = read_qstring(fid)
                (
                    new_channel["native_order"],
                    new_channel["custom_order"],
                    signal_type,
                    channel_enabled,
                    new_channel["chip_channel"],
                    command_stream,
                    new_channel["board_stream"],
                ) = struct.unpack(
                    "<hhhhhhh", fid.read(14)
                )  # ignore command_stream
                new_trigger_channel = {}
                (
                    new_trigger_channel["voltage_trigger_mode"],
                    new_trigger_channel["voltage_threshold"],
                    new_trigger_channel["digital_trigger_channel"],
                    new_trigger_channel["digital_edge_polarity"],
                ) = struct.unpack("<hhhh", fid.read(8))
                (
                    new_channel["electrode_impedance_magnitude"],
                    new_channel["electrode_impedance_phase"],
                ) = struct.unpack("<ff", fid.read(8))

                if channel_enabled:
                    if signal_type == 0:
                        header["amplifier_channels"].append(new_channel)

                        # If dc amplifier data is being saved, dc_amplifier_channels
                        if header["dc_amplifier_data_saved"]:
                            new_dc_channel = {
                                "port_name": new_channel["port_name"],
                                "port_prefix": new_channel["port_prefix"],
                                "port_number": new_channel["port_number"],
                                "native_channel_name": "DC_"
                                + new_channel["native_channel_name"],
                                "custom_channel_name": "DC_"
                                + new_channel["custom_channel_name"],
                                "native_order": new_channel["native_order"],
                                "custom_order": new_channel["custom_order"],
                                "chip_channel": new_channel["chip_channel"],
                                "board_stream": new_channel["board_stream"],
                                "electrode_impedance_magnitude": new_channel[
                                    "electrode_impedance_magnitude"
                                ],
                                "electrode_impedance_phase": new_channel[
                                    "electrode_impedance_phase"
                                ],
                            }
                            header["dc_amplifier_channels"].append(new_dc_channel)

                        # stim_channels
                        new_stim_channel = {
                            "port_name": new_channel["port_name"],
                            "port_prefix": new_channel["port_prefix"],
                            "port_number": new_channel["port_number"],
                            "native_channel_name": "STIM_"
                            + new_channel["native_channel_name"],
                            "custom_channel_name": "STIM_"
                            + new_channel["custom_channel_name"],
                            "native_order": new_channel["native_order"],
                            "custom_order": new_channel["custom_order"],
                            "chip_channel": new_channel["chip_channel"],
                            "board_stream": new_channel["board_stream"],
                            "electrode_impedance_magnitude": new_channel[
                                "electrode_impedance_magnitude"
                            ],
                            "electrode_impedance_phase": new_channel[
                                "electrode_impedance_phase"
                            ],
                        }
                        header["stim_channels"].append(new_stim_channel)

                        # amp_settle_channels
                        new_amp_settle_channel = {
                            "port_name": new_channel["port_name"],
                            "port_prefix": new_channel["port_prefix"],
                            "port_number": new_channel["port_number"],
                            "native_channel_name": "AMP_SETTLE_"
                            + new_channel["native_channel_name"],
                            "custom_channel_name": "AMP_SETTLE_"
                            + new_channel["custom_channel_name"],
                            "native_order": new_channel["native_order"],
                            "custom_order": new_channel["custom_order"],
                            "chip_channel": new_channel["chip_channel"],
                            "board_stream": new_channel["board_stream"],
                            "electrode_impedance_magnitude": new_channel[
                                "electrode_impedance_magnitude"
                            ],
                            "electrode_impedance_phase": new_channel[
                                "electrode_impedance_phase"
                            ],
                        }
                        header["amp_settle_channels"].append(new_amp_settle_channel)

                        # charge_recovery_channels
                        new_charge_recovery_channel = {
                            "port_name": new_channel["port_name"],
                            "port_prefix": new_channel["port_prefix"],
                            "port_number": new_channel["port_number"],
                            "native_channel_name": "CHARGE_RECOVERY_"
                            + new_channel["native_channel_name"],
                            "custom_channel_name": "CHARGE_RECOVERY_"
                            + new_channel["custom_channel_name"],
                            "native_order": new_channel["native_order"],
                            "custom_order": new_channel["custom_order"],
                            "chip_channel": new_channel["chip_channel"],
                            "board_stream": new_channel["board_stream"],
                            "electrode_impedance_magnitude": new_channel[
                                "electrode_impedance_magnitude"
                            ],
                            "electrode_impedance_phase": new_channel[
                                "electrode_impedance_phase"
                            ],
                        }
                        header["charge_recovery_channels"].append(
                            new_charge_recovery_channel
                        )

                        # compliance_limit_channels
                        new_compliance_limit_channel = {
                            "port_name": new_channel["port_name"],
                            "port_prefix": new_channel["port_prefix"],
                            "port_number": new_channel["port_number"],
                            "native_channel_name": "COMPLIANCE_LIMIT_"
                            + new_channel["native_channel_name"],
                            "custom_channel_name": "COMPLIANCE_LIMIT_"
                            + new_channel["custom_channel_name"],
                            "native_order": new_channel["native_order"],
                            "custom_order": new_channel["custom_order"],
                            "chip_channel": new_channel["chip_channel"],
                            "board_stream": new_channel["board_stream"],
                            "electrode_impedance_magnitude": new_channel[
                                "electrode_impedance_magnitude"
                            ],
                            "electrode_impedance_phase": new_channel[
                                "electrode_impedance_phase"
                            ],
                        }
                        header["compliance_limit_channels"].append(
                            new_compliance_limit_channel
                        )

                        header["spike_triggers"].append(new_trigger_channel)
                    elif signal_type == 1:
                        raise Exception("Wrong signal type for the rhs format")
                        # header['aux_input_channels'].append(new_channel)
                    elif signal_type == 2:
                        raise Exception("Wrong signal type for the rhs format")
                        # header['supply_voltage_channels'].append(new_channel)
                    elif signal_type == 3:
                        header["board_adc_channels"].append(new_channel)
                    elif signal_type == 4:
                        header["board_dac_channels"].append(new_channel)
                    elif signal_type == 5:
                        header["board_dig_in_channels"].append(new_channel)
                    elif signal_type == 6:
                        header["board_dig_out_channels"].append(new_channel)
                    else:
                        raise Exception("Unknown channel type.")

    # Summarize contents of data file.
    header["num_amplifier_channels"] = len(header["amplifier_channels"])
    header["num_board_adc_channels"] = len(header["board_adc_channels"])
    header["num_board_dac_channels"] = len(header["board_dac_channels"])
    header["num_board_dig_in_channels"] = len(header["board_dig_in_channels"])
    header["num_board_dig_out_channels"] = len(header["board_dig_out_channels"])

    return header


# Define notch_filter function
def notch_filter(input, f_sample, f_notch, bandwidth):  # pragma: no cover
    """Implements a notch filter (e.g., for 50 or 60 Hz) on vector 'input'.

    f_sample = sample rate of data (input Hz or Samples/sec)
    f_notch = filter notch frequency (input Hz)
    bandwidth = notch 3-dB bandwidth (input Hz).  A bandwidth of 10 Hz is
    recommended for 50 or 60 Hz notch filters; narrower bandwidths lead to
    poor time-domain properties with an extended ringing response to
    transient disturbances.

    Example:  If neural data was sampled at 30 kSamples/sec
    and you wish to implement a 60 Hz notch filter:

    output = notch_filter(input, 30000, 60, 10);
    """
    t_step = 1.0 / f_sample
    f_c = f_notch * t_step

    L = len(input)

    # Calculate IIR filter parameters
    d = math.exp(-2.0 * math.pi * (bandwidth / 2.0) * t_step)
    b = (1.0 + d * d) * math.cos(2.0 * math.pi * f_c)
    a0 = 1.0
    a1 = -b
    a2 = d * d
    a = (1.0 + d * d) / 2.0
    b0 = 1.0
    b1 = -2.0 * math.cos(2.0 * math.pi * f_c)
    b2 = 1.0

    output = np.zeros(len(input))
    output[0] = input[0]
    output[1] = input[1]
    # (If filtering a continuous data stream, change output[0:1] to the
    #  previous final two values of output.)

    # Run filter
    for i in range(2, L):
        output[i] = (
            a * b2 * input[i - 2]
            + a * b1 * input[i - 1]
            + a * b0 * input[i]
            - a2 * output[i - 2]
            - a1 * output[i - 1]
        ) / a0

    return output


# Define find_channel_in_group function
def find_channel_in_group(channel_name, signal_group):  # pragma: no cover
    for count, this_channel in enumerate(signal_group):
        if this_channel["custom_channel_name"] == channel_name:
            return True, count
    return False, 0


# Define find_channel_in_header function
def find_channel_in_header(channel_name, header):  # pragma: no cover

    # Look through all present signal groups

    # 1. Look through amplifier_channels
    if "amplifier_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["amplifier_channels"]
        )
        if channel_found:
            return True, "amplifier_channels", channel_index

    # 2. Look through dc_amplifier_channels
    if "dc_amplifier_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["dc_amplifier_channels"]
        )
        if channel_found:
            return True, "dc_amplifier_channels", channel_index

    # 3. Look through stim_channels
    if "stim_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["stim_channels"]
        )
        if channel_found:
            return True, "stim_channels", channel_index

    # 3.1 Look through amp_settle_channels
    if "amp_settle_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["amp_settle_channels"]
        )
        if channel_found:
            return True, "amp_settle_channels", channel_index

    # 3.2 Look through charge_recovery_channels
    if "charge_recovery_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["charge_recovery_channels"]
        )
        if channel_found:
            return True, "charge_recovery_channels", channel_index

    # 3.3 Look through compliance_limit_channels
    if "compliance_limit_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["compliance_limit_channels"]
        )
        if channel_found:
            return True, "compliance_limit_channels", channel_index

    # 4. Look through board_adc_channels
    if "board_adc_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["board_adc_channels"]
        )
        if channel_found:
            return True, "board_adc_channels", channel_index

    # 5. Look through board_dac_channels
    if "board_dac_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["board_dac_channels"]
        )
        if channel_found:
            return True, "board_dac_channels", channel_index

    # 6. Look through board_dig_in_channels
    if "board_dig_in_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["board_dig_in_channels"]
        )
        if channel_found:
            return True, "board_dig_in_channels", channel_index

    # 7. Look through board_dig_out_channels
    if "board_dig_out_channels" in header:
        channel_found, channel_index = find_channel_in_group(
            channel_name, header["board_dig_out_channels"]
        )
        if channel_found:
            return True, "board_dig_out_channels", channel_index

    return False, "", 0


# Define get_bytes_per_data_block function
def get_bytes_per_data_block(header):  # pragma: no cover
    """Calculates the number of bytes in each 128-sample datablock."""
    N = 128  # n of amplifier samples
    # Each data block contains N amplifier samples.
    bytes_per_block = N * 4  # timestamp data

    bytes_per_block += N * 2 * header["num_amplifier_channels"]

    # DC amplifier voltage (absent if flag was off)
    # bytes_per_block += N * 2 * header['dc_amplifier_data_saved']
    if header["dc_amplifier_data_saved"] > 0:
        bytes_per_block += N * 2 * header["num_amplifier_channels"]

    # Stimulation data, one per enabled amplifier channels
    bytes_per_block += N * 2 * header["num_amplifier_channels"]

    # Board analog inputs are sampled at same rate as amplifiers
    bytes_per_block += N * 2 * header["num_board_adc_channels"]

    # Board analog outputs are sampled at same rate as amplifiers
    bytes_per_block += N * 2 * header["num_board_dac_channels"]

    # Board digital inputs are sampled at same rate as amplifiers
    if header["num_board_dig_in_channels"] > 0:
        bytes_per_block += N * 2

    # Board digital outputs are sampled at same rate as amplifiers
    if header["num_board_dig_out_channels"] > 0:
        bytes_per_block += N * 2

    return bytes_per_block


# Define read_one_data_block function
def read_one_data_block(data, header, indices, fid):  # pragma: no cover
    """Reads one 128-sample data block from fid into data, at the location indicated by indices."""

    data["t"][indices["amplifier"] : (indices["amplifier"] + 128)] = np.array(
        struct.unpack("<" + "i" * 128, fid.read(128 * 4))
    )

    if header["num_amplifier_channels"] > 0:
        tmp = np.fromfile(
            fid, dtype="uint16", count=128 * header["num_amplifier_channels"]
        )
        data["amplifier_data"][
            range(header["num_amplifier_channels"]),
            indices["amplifier"] : (indices["amplifier"] + 128),
        ] = tmp.reshape(header["num_amplifier_channels"], 128)

        # check if dc amplifier voltage was saved
        if header["dc_amplifier_data_saved"]:
            tmp = np.fromfile(
                fid, dtype="uint16", count=128 * header["num_amplifier_channels"]
            )
            data["dc_amplifier_data"][
                range(header["num_amplifier_channels"]),
                indices["amplifier"] : (indices["amplifier"] + 128),
            ] = tmp.reshape(header["num_amplifier_channels"], 128)

        # get the stimulation data
        tmp = np.fromfile(
            fid, dtype="uint16", count=128 * header["num_amplifier_channels"]
        )
        data["stim_data_raw"][
            range(header["num_amplifier_channels"]),
            indices["amplifier"] : (indices["amplifier"] + 128),
        ] = tmp.reshape(header["num_amplifier_channels"], 128)

    if header["num_board_adc_channels"] > 0:
        tmp = np.fromfile(
            fid, dtype="uint16", count=128 * header["num_board_adc_channels"]
        )
        data["board_adc_data"][
            range(header["num_board_adc_channels"]),
            indices["board_adc"] : (indices["board_adc"] + 128),
        ] = tmp.reshape(header["num_board_adc_channels"], 128)

    if header["num_board_dac_channels"] > 0:
        tmp = np.fromfile(
            fid, dtype="uint16", count=128 * header["num_board_dac_channels"]
        )
        data["board_dac_data"][
            range(header["num_board_dac_channels"]),
            indices["board_dac"] : (indices["board_dac"] + 128),
        ] = tmp.reshape(header["num_board_dac_channels"], 128)

    if header["num_board_dig_in_channels"] > 0:
        data["board_dig_in_raw"][
            indices["board_dig_in"] : (indices["board_dig_in"] + 128)
        ] = np.array(struct.unpack("<" + "H" * 128, fid.read(256)))

    if header["num_board_dig_out_channels"] > 0:
        data["board_dig_out_raw"][
            indices["board_dig_out"] : (indices["board_dig_out"] + 128)
        ] = np.array(struct.unpack("<" + "H" * 128, fid.read(256)))


# Define data_to_result function
def data_to_result(header, data, data_present):  # pragma: no cover
    """Moves the header and data (if present) into a common object."""

    result = {}

    stim_parameters = {}
    stim_parameters["stim_step_size"] = header["stim_step_size"]
    stim_parameters["charge_recovery_current_limit"] = header["recovery_current_limit"]
    stim_parameters["charge_recovery_target_voltage"] = header[
        "recovery_target_voltage"
    ]
    stim_parameters["amp_settle_mode"] = header["amp_settle_mode"]
    stim_parameters["charge_recovery_mode"] = header["charge_recovery_mode"]
    result["stim_parameters"] = stim_parameters

    result["stim_channels"] = header["stim_channels"]
    result["spike_triggers"] = header["spike_triggers"]
    result["notes"] = header["notes"]
    result["frequency_parameters"] = header["frequency_parameters"]

    result["reference_channel"] = header["reference_channel"]
    result["amplifier_channels"] = header["amplifier_channels"]
    result["board_adc_channels"] = header["board_adc_channels"]
    result["board_dac_channels"] = header["board_dac_channels"]

    result["dc_amplifier_data_saved"] = header["dc_amplifier_data_saved"]
    if header["dc_amplifier_data_saved"]:
        result["dc_amplifier_channels"] = header["dc_amplifier_channels"]

    result["compliance_limit_channels"] = header["compliance_limit_channels"]
    result["charge_recovery_channels"] = header["charge_recovery_channels"]
    result["amp_settle_channels"] = header["amp_settle_channels"]
    result["board_dig_in_channels"] = header["board_dig_in_channels"]
    result["board_dig_out_channels"] = header["board_dig_out_channels"]

    if data_present > 0:
        result["t"] = data["t"]
        result["stim_data"] = data["stim_data"]
        result["amplifier_data"] = data["amplifier_data"]
        result["board_adc_data"] = data["board_adc_data"]
        result["board_dac_data"] = data["board_dac_data"]
        if header["dc_amplifier_data_saved"] > 0:
            result["dc_amplifier_data"] = data["dc_amplifier_data"]
        result["compliance_limit_data"] = data["compliance_limit_data"]
        result["charge_recovery_data"] = data["charge_recovery_data"]
        result["amp_settle_data"] = data["amp_settle_data"]
        result["board_dig_in_data"] = data["board_dig_in_data"]
        result["board_dig_out_data"] = data["board_dig_out_data"]

    return result


# Define plot_channel function
def plot_channel(channel_name, result):  # pragma: no cover
    # Find channel that corresponds to this name
    channel_found, signal_type, signal_index = find_channel_in_header(
        channel_name, result
    )

    # Plot this channel
    if channel_found:
        fig, ax = plt.subplots()
        ax.set_title(channel_name)
        ax.set_xlabel("Time (s)")

        if signal_type == "amplifier_channels":
            ylabel = "Voltage (microVolts)"
            signal_data_name = "amplifier_data"

        elif signal_type == "dc_amplifier_channels":
            ylabel = "Voltage (Volts)"
            signal_data_name = "dc_amplifier_data"

        elif signal_type == "stim_channels":
            ylabel = "Current (microAmps)"
            signal_data_name = "stim_data"

        elif signal_type == "amp_settle_channels":
            ylabel = "Amp Settle Events (High or Low)"
            signal_data_name = "amp_settle_data"

        elif signal_type == "charge_recovery_channels":
            ylabel = "Charge Recovery Events (High or Low)"
            signal_data_name = "charge_recovery_data"

        elif signal_type == "compliance_limit_channels":
            ylabel = "Compliance Limit Events (High or Low)"
            signal_data_name = "compliance_limit_data"

        elif signal_type == "board_adc_channels":
            ylabel = "Voltage (Volts)"
            signal_data_name = "board_adc_data"

        elif signal_type == "board_dac_channels":
            ylabel = "Voltage (Volts)"
            signal_data_name = "board_dac_data"

        elif signal_type == "board_dig_in_channels":
            ylabel = "Digital In Events (High or Low)"
            signal_data_name = "board_dig_in_data"

        elif signal_type == "board_dig_out_channels":
            ylabel = "Digital Out Events (High or Low)"
            signal_data_name = "board_dig_out_data"

        else:
            raise Exception(
                "Plotting not possible; signal type ", signal_type, " not found"
            )

        ax.set_ylabel(ylabel)

        ax.plot(result["t"], result[signal_data_name][signal_index, :])
        ax.margins(x=0, y=0)

    else:
        raise Exception("Plotting not possible; channel ", channel_name, " not found")


# Define load_file function
def load_file(filename):  # pragma: no cover
    # Start timing
    tic = time.time()

    # Open file
    fid = open(filename, "rb")
    filesize = os.path.getsize(filename)

    # Read file header
    header = read_header(fid)

    # Output a summary of recorded data
    logging.info(
        "Found {} amplifier channel{}.".format(
            header["num_amplifier_channels"], plural(header["num_amplifier_channels"])
        )
    )
    logging.info(
        "Found {} board ADC channel{}.".format(
            header["num_board_adc_channels"], plural(header["num_board_adc_channels"])
        )
    )
    logging.info(
        "Found {} board DAC channel{}.".format(
            header["num_board_dac_channels"], plural(header["num_board_dac_channels"])
        )
    )
    logging.info(
        "Found {} board digital input channel{}.".format(
            header["num_board_dig_in_channels"],
            plural(header["num_board_dig_in_channels"]),
        )
    )
    logging.info(
        "Found {} board digital output channel{}.".format(
            header["num_board_dig_out_channels"],
            plural(header["num_board_dig_out_channels"]),
        )
    )
    logging.info("")

    # Determine how many samples the data file contains
    bytes_per_block = get_bytes_per_data_block(header)
    logging.info(f"{bytes_per_block} bytes per data block")
    # How many data blocks remain in this file?
    data_present = False
    bytes_remaining = filesize - fid.tell()
    if bytes_remaining > 0:
        data_present = True

    if bytes_remaining % bytes_per_block != 0:
        raise Exception(
            "Something is wrong with file size : should have a whole number of data blocks"
        )

    num_data_blocks = int(bytes_remaining / bytes_per_block)

    # Calculate how many samples of each signal type are present
    num_amplifier_samples = 128 * num_data_blocks
    num_board_adc_samples = 128 * num_data_blocks
    num_board_dac_samples = 128 * num_data_blocks
    num_board_dig_in_samples = 128 * num_data_blocks
    num_board_dig_out_samples = 128 * num_data_blocks

    record_time = num_amplifier_samples / header["sample_rate"]

    # Output a summary of contents of header file
    if data_present:
        logging.info(
            "File contains {:0.3f} seconds of data.  Amplifiers were sampled at {:0.2f} kS/s.".format(
                record_time, header["sample_rate"] / 1000
            )
        )
    else:
        logging.info(
            "Header file contains no data.  Amplifiers were sampled at {:0.2f} kS/s.".format(
                header["sample_rate"] / 1000
            )
        )

    if data_present:
        # Pre-allocate memory for data
        logging.info("")
        logging.info("Allocating memory for data...")

        data = {}
        data["t"] = np.zeros(num_amplifier_samples, dtype=np.int_)

        data["amplifier_data"] = np.zeros(
            [header["num_amplifier_channels"], num_amplifier_samples], dtype=np.uint
        )

        if header["dc_amplifier_data_saved"]:
            data["dc_amplifier_data"] = (
                np.zeros(
                    [header["num_amplifier_channels"], num_amplifier_samples],
                    dtype=np.uint,
                )
                * header["dc_amplifier_data_saved"]
            )

        data["stim_data_raw"] = np.zeros(
            [header["num_amplifier_channels"], num_amplifier_samples], dtype=np.int_
        )
        data["stim_data"] = np.zeros(
            [header["num_amplifier_channels"], num_amplifier_samples], dtype=np.int_
        )

        data["board_adc_data"] = np.zeros(
            [header["num_board_adc_channels"], num_board_adc_samples], dtype=np.uint
        )
        data["board_dac_data"] = np.zeros(
            [header["num_board_dac_channels"], num_board_dac_samples], dtype=np.uint
        )

        # by default, this script interprets digital events (digital inputs and outputs) as booleans
        # if unsigned int values are preferred(0 for False, 1 for True), replace the 'dtype=np.bool' argument with 'dtype=np.uint' as shown
        # the commented line below illustrates this for digital input data; the same can be done for digital out

        # data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples], dtype=np.uint)
        data["board_dig_in_data"] = np.zeros(
            [header["num_board_dig_in_channels"], num_board_dig_in_samples],
            dtype=np.bool_,
        )
        data["board_dig_in_raw"] = np.zeros(num_board_dig_in_samples, dtype=np.uint)
        data["board_dig_out_data"] = np.zeros(
            [header["num_board_dig_out_channels"], num_board_dig_out_samples],
            dtype=np.bool_,
        )
        data["board_dig_out_raw"] = np.zeros(num_board_dig_out_samples, dtype=np.uint)

        # Read sampled data from file.
        logging.info("Reading data from file...")

        # Initialize indices used in looping
        indices = {}
        indices["amplifier"] = 0
        indices["board_adc"] = 0
        indices["board_dac"] = 0
        indices["board_dig_in"] = 0
        indices["board_dig_out"] = 0

        print_increment = 10
        percent_done = print_increment
        for i in range(num_data_blocks):
            read_one_data_block(data, header, indices, fid)
            # Increment all indices in 128
            indices = {k: v + 128 for k, v in indices.items()}

            fraction_done = 100 * (1.0 * i / num_data_blocks)
            if fraction_done >= percent_done:
                logging.info(f"{percent_done}% done...")
                percent_done = percent_done + print_increment

        logging.info("100% done...")

        # Make sure we have read exactly the right amount of data
        bytes_remaining = filesize - fid.tell()
        if bytes_remaining != 0:
            raise Exception("Error: End of file not reached.")

    fid.close()

    if data_present:
        logging.info("Parsing data...")

        # Extract digital input channels to separate variables.
        for i in range(header["num_board_dig_in_channels"]):
            data["board_dig_in_data"][i, :] = np.not_equal(
                np.bitwise_and(
                    data["board_dig_in_raw"],
                    (1 << header["board_dig_in_channels"][i]["native_order"]),
                ),
                0,
            )

        # Extract digital output channels to separate variables.
        for i in range(header["num_board_dig_out_channels"]):
            data["board_dig_out_data"][i, :] = np.not_equal(
                np.bitwise_and(
                    data["board_dig_out_raw"],
                    (1 << header["board_dig_out_channels"][i]["native_order"]),
                ),
                0,
            )

        # Extract stimulation data
        data["compliance_limit_data"] = (
            np.bitwise_and(data["stim_data_raw"], 32768) >= 1
        )  # get 2^15 bit, interpret as True or False
        data["charge_recovery_data"] = (
            np.bitwise_and(data["stim_data_raw"], 16384) >= 1
        )  # get 2^14 bit, interpret as True or False
        data["amp_settle_data"] = (
            np.bitwise_and(data["stim_data_raw"], 8192) >= 1
        )  # get 2^13 bit, interpret as True or False
        data["stim_polarity"] = 1 - (
            2 * (np.bitwise_and(data["stim_data_raw"], 256) >> 8)
        )  # get 2^8 bit, interpret as +1 for 0_bit or -1 for 1_bit

        curr_amp = np.bitwise_and(
            data["stim_data_raw"], 255
        )  # get least-significant 8 bits corresponding to the current amplitude
        data["stim_data"] = (
            curr_amp * data["stim_polarity"]
        )  # multiply current amplitude by the correct sign

        # # Scale voltage levels appropriately.
        data["amplifier_data"] = np.multiply(
            0.195, (data["amplifier_data"].astype(np.int32) - 32768)
        )  # units = microvolts
        data["stim_data"] = np.multiply(
            header["stim_step_size"], data["stim_data"] / 1.0e-6
        )

        if header["dc_amplifier_data_saved"]:
            data["dc_amplifier_data"] = np.multiply(
                -0.01923, (data["dc_amplifier_data"].astype(np.int32) - 512)
            )  # units = volts

        data["board_adc_data"] = np.multiply(
            0.0003125, (data["board_adc_data"].astype(np.int32) - 32768)
        )  # units = volts
        data["board_dac_data"] = np.multiply(
            0.0003125, (data["board_dac_data"].astype(np.int32) - 32768)
        )  # units = volts

        # Check for gaps in timestamps.
        num_gaps = np.sum(np.not_equal(data["t"][1:] - data["t"][:-1], 1))
        if num_gaps == 0:
            logging.info("No missing timestamps in data.")
        else:
            logging.info(
                f"Warning: {num_gaps} gaps in timestamp data found.  Time scale will not be uniform!"
            )

        # Scale time steps (units = seconds).
        data["t"] = data["t"] / header["sample_rate"]

        # If the software notch filter was selected during the recording, apply the
        # same notch filter to amplifier data here.
        if header["notch_filter_frequency"] > 0 and header["version"]["major"] < 3:
            print_increment = 10
            percent_done = print_increment
            for i in range(header["num_amplifier_channels"]):
                data["amplifier_data"][i, :] = notch_filter(
                    data["amplifier_data"][i, :],
                    header["sample_rate"],
                    header["notch_filter_frequency"],
                    10,
                )
                if fraction_done >= percent_done:
                    logging.info(f"{percent_done}% done...")
                    percent_done = percent_done + print_increment
    else:
        data = []

    # Move variables to result struct.
    result = data_to_result(header, data, data_present)

    logging.info(f"Done!  Elapsed time: {time.time() - tic:0.1f} seconds")

    return result, data_present


# Define function print_all_channel_names
def print_all_channel_names(result):  # pragma: no cover

    # Print all amplifier_channels
    if "amplifier_channels" in result:
        print_names_in_group(result["amplifier_channels"])

    # Print all dc_amplifier_channels
    if "dc_amplifier_channels" in result:
        print_names_in_group(result["dc_amplifier_channels"])

    # Print all stim_channels
    if "stim_channels" in result:
        print_names_in_group(result["stim_channels"])

    # Print all amp_settle_channels
    if "amp_settle_channels" in result:
        print_names_in_group(result["amp_settle_channels"])

    # Print all charge_recovery_channels
    if "charge_recovery_channels" in result:
        print_names_in_group(result["charge_recovery_channels"])

    # Print all compliance_limit_channels
    if "compliance_limit_channels" in result:
        print_names_in_group(result["compliance_limit_channels"])

    # Print all board_adc_channels
    if "board_adc_channels" in result:
        print_names_in_group(result["board_adc_channels"])

    # Print all board_dac_channels
    if "board_dac_channels" in result:
        print_names_in_group(result["board_dac_channels"])

    # Print all board_dig_in_channels
    if "board_dig_in_channels" in result:
        print_names_in_group(result["board_dig_in_channels"])

    # Print all board_dig_out_channels
    if "board_dig_out_channels" in result:
        print_names_in_group(result["board_dig_out_channels"])


# Define function print_names_in_group
def print_names_in_group(signal_group):  # pragma: no cover
    for this_channel in signal_group:
        logging.info(this_channel["custom_channel_name"])



# Pairwise Granger Causality
import os

import numpy as np
from elephant.causality.granger import pairwise_granger

from miv.typing import SignalType


def pairwise_causality(signal: SignalType, start: int, end: int):
    """
    Estimates pairwise Granger Causality between all channels.

    Parameters
    ----------
    signal : SignalType
       Input signal.
    start : int
       starting point from signal
    end : int
       End point from signal

    Returns
    -------
    C : np.ndarray
        Causality Matrix (shape=2x2) containing directional causalities for X -> Y and Y -> X,
        instantaneous causality between X,Y, and total causality. X and Y represents electrodes

    See Also
    --------
    miv.visualization.causality.pairwise_causality_plot

    """

    p = len(signal[0])  # Number of channels
    C = np.zeros((4, p, p))  # Causality Matrix

    for j in range(p):
        for k in range(j + 1, p):
            C[:, j, k] = pairwise_granger(
                np.transpose([signal[start:end, j], signal[start:end, k]]),
                max_order=1,
            )
            C[:, k, j] = pairwise_granger(
                np.transpose([signal[start:end, k], signal[start:end, j]]),
                max_order=1,
            )
    # for i in range(4):
    #    np.fill_diagonal(C[i], 0.0)

    # C or causality matrix contains four p X p matrices. These are directional causalities for X -> Y and Y -> X,
    # instantaneous causality between X,Y, and total causality. X and Y represents electrodes
    return C



__all__ = [
    "firing_rates",
    "interspike_intervals",
    "coefficient_variation",
    "peri_stimulus_time",
    "binned_spiketrain",
    "fano_factor",
    "decay_spike_counts",
    "spike_counts_with_kernel",
]

from typing import Any, Dict, Iterable, List, Optional, Union

import datetime

import elephant.statistics
import matplotlib.pyplot as plt
import neo
import numpy as np
import quantities as pq
import scipy
import scipy.signal

from miv.typing import SpikestampsType


# FIXME: For now, we provide the free function for simple usage. For more
# advanced statistical analysis, we should have a module wrapper.
def firing_rates(
    spiketrains: Union[pq.Quantity, Iterable[neo.core.SpikeTrain]],
    # t_start: Optional[float] = None,
    # t_stop: Optional[float] = None,
) -> Dict[str, Any]:
    """
    Process basic spiketrains statistics: rates, mean, variance.

    Parameters
    ----------
    spiketrains : Iterable[neo.core.SpikeTrain]

    Returns
    -------
    Iterable[Any]

    """
    rates = []
    for spikestamp in spiketrains:
        if len(spikestamp) == 0:
            rates.append(0)
            continue
        mfr = elephant.statistics.mean_firing_rate(spikestamp)
        if isinstance(mfr, pq.quantity.Quantity):
            mfr = mfr.magnitude
        rates.append(mfr)

    rates_mean_over_channel = np.mean(rates)
    rates_variance_over_channel = np.var(rates)
    return {
        "rates": rates,
        "mean": rates_mean_over_channel,
        "variance": rates_variance_over_channel,
    }


def interspike_intervals(spikes: SpikestampsType):
    """
    Compute the inter-spike intervals of the given spike train.

    Examples
    --------

    How to draw Inter-spike interval histogram (ISIH):

        >>> from miv.statistics import interspike_intervals
        >>> import matplotlib.pyplot as plt
        >>> interspike_interval = interspike_intervals(spikestamp)
        >>> plt.hist(interspike_interval)

    If one wants to get the bin-count, you can use `numpy.digitize` and `numpy.bincount`:

        >>> import numpy as np
        >>> max_time = spikestamps.max()
        >>> num_bins = 20
        >>> time_interval = np.linspace(0, max_time, num_bins)
        >>> digitized = np.digitize(interspike_interval, time_interval)
        >>> bin_counts = np.bincount(digitized)

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        interval: numpy.ndarray

    """

    spike_interval = np.diff(spikes)
    return spike_interval


def coefficient_variation(self, spikes: SpikestampsType):
    """
    The Coefficient of variation: ratio of interspike standard deviation over mean.

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        coefficient variation : float
    """
    interspike = self.interspike_intervals()
    return np.std(interspike) / np.mean(interspike)


def peri_stimulus_time(spike_list: List[SpikestampsType]):
    """
    Compute the peri-stimulus time of the given spike train.

    Examples
    --------

    How to draw Peri-stimulus time histogram (ISIH):

        >>> from miv.statistics import peri_stimulus_time
        >>> import matplotlib.pyplot as plt
        >>> pst = peri_stimulus_time(spikestamp)
        >>> plt.hist(pst)

    If one wants to get the bin-count, you can use `numpy.digitize` and `numpy.bincount`:

        >>> import numpy as np
        >>> max_time = spikestamps.max()
        >>> num_bins = 20
        >>> time_interval = np.linspace(0, max_time, num_bins)
        >>> digitized = np.digitize(pst, time_interval)
        >>> bin_counts = np.bincount(digitized)

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        interval: numpy.ndarray

    """

    peri_stimulus_times = np.sum(np.array(spike_list), 0)
    return peri_stimulus_times


def binned_spiketrain(
    spiketrain: SpikestampsType,
    t_start: float,
    t_end: float,
    bin_size: float,
    return_count: bool = False,
):
    """
    Forms a binned spiketrain using the spiketrain

    Parameters
    ----------
    spiketrain : SpikestampsType
        Single spike-stamp
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds
    return_count : bool
        If set to true, return the bin count. (default=False)

    Returns
    -------
        bin_spike: numpy.ndarray
        binned spiketrain with 1 corresponding to spike and zero otherwise

    """
    assert t_start < t_end, "End time cannot be smaller or equal to start time"
    assert bin_size > 0, "bin size should be greater than 0"
    n_bins = int(np.ceil((t_end - t_start) / bin_size))
    time = t_start + (np.arange(n_bins + 1) * bin_size)
    if isinstance(spiketrain, neo.core.SpikeTrain):
        bins = np.digitize(spiketrain.magnitude, time)
    else:
        bins = np.digitize(spiketrain, time)
    bincount = np.bincount(bins, minlength=n_bins + 2)[1:-1]
    if return_count:
        bin_spike = bincount
    else:
        bin_spike = (bincount != 0).astype(np.int_)

    return bin_spike


def fano_factor(
    spiketrains: SpikestampsType,
    channel: float,  # TODO: the function should be independent of channel.
    t_start: float,
    t_end: float,
    n_bins: float,
):
    """
    Calculates the Fano factor for given signal by dividing it into the specified number of bins

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channel : float
        electrode/channel
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    n_bins : float
        Number of bins

    Returns
    -------
        fano_fac: float
        fanofactor for the specified channel and conditions

    """
    assert (
        t_start < t_end
    ), f"End time {t_end} cannot be smaller or equal to start time {t_start}."
    assert n_bins > 0, "Number of bins should be a positive integer"
    bin_spike = binned_spiketrain(spiketrains[channel], t_start, t_end, 0.002)
    assert np.sum(bin_spike) != 0, "The channel has no spikes"
    large_bin = []
    bin_length = np.int32(np.size(bin_spike) / n_bins)
    count = 0
    for i in np.arange(n_bins):
        large_bin.append(np.sum(bin_spike[count : count + bin_length]))
        count += bin_length
    bin_array = np.array(large_bin)
    fano_fac = np.var(bin_array) / np.mean(bin_array)

    return fano_fac


def decay_spike_counts(
    spiketrain, probe_times, amplitude=1.0, decay_rate=5, batchsize=256
):
    """
    decay_spike_counts.

    Both spiketrain and probe_times should be a 1-d array representing time.

    Parameters
    ----------
    spiketrain :
        spiketrain
    probe_times :
        probe_times
    amplitude :
        amplitude
    decay_rate :
        decay_rate
    batchsize :
        batchsize
    """
    if len(spiketrain) == 0:
        return np.zeros_like(probe_times)
    batchsize = min(spiketrain.shape[0], batchsize)
    num_sections = spiketrain.shape[0] // batchsize + (
        1 if spiketrain.shape[0] % batchsize > 0 else 0
    )

    result = np.zeros_like(probe_times)
    for subspiketrain in np.array_split(spiketrain, num_sections):
        exponent = (
            np.tile(probe_times, (len(subspiketrain), 1)) - subspiketrain[:, None]
        )
        mask = exponent < 0
        exponent[mask] = 0
        decay_count = np.exp(-decay_rate * exponent)
        decay_count[mask] = 0
        result += decay_count.sum(axis=0)
    return result


def spike_counts_with_kernel(spiketrain, probe_times, kernel, batchsize=32):
    if len(spiketrain) == 0:
        return np.zeros_like(probe_times)

    batchsize = min(spiketrain.shape[0], batchsize)
    num_sections = spiketrain.shape[0] // batchsize + (
        1 if spiketrain.shape[0] % batchsize > 0 else 0
    )

    result = np.zeros_like(probe_times)
    for subspiketrain in np.array_split(spiketrain, num_sections):
        exponent = (
            np.tile(probe_times, (len(subspiketrain), 1)) - subspiketrain[:, None]
        )
        mask = exponent < 0
        exponent[mask] = 0
        decay_count = kernel(exponent)
        # decay_count = np.exp(-decay_rate * exponent)
        decay_count[mask] = 0
        result += decay_count.sum(axis=0)
    return result



#! /usr/bin/env python

# This file is part of the SpikeCoding repository - MAVLab TU Delft
#
# MIT License
#
# Copyright (c) 2021 Julien Dupeyroux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# @author Julien Dupeyroux, Stein Stroobants

import numpy as np


def temporal_contrast(spikes, threshold):
    # Based on algorithm provided in:
    #   Sengupta et al. (2017)
    #   Petro et al. (2020)
    signal = np.zeros(len(spikes))
    for i in range(1, len(spikes)):
        if spikes[i] > 0:
            signal[i] = signal[i - 1] + threshold
        elif spikes[i] < 0:
            signal[i] = signal[i - 1] - threshold
        else:
            signal[i] = signal[i - 1]
    return signal


def step_forward(spikes, threshold, startpoint):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    signal = np.zeros(len(spikes))
    signal[0] = startpoint
    for i in range(1, len(spikes)):
        if spikes[i] > 0:
            signal[i] = signal[i - 1] + threshold
        elif spikes[i] < 0:
            signal[i] = signal[i - 1] - threshold
        else:
            signal[i] = signal[i - 1]
    return signal


def moving_window(spikes, threshold, startpoint):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    signal = np.zeros(len(spikes))
    signal[0] = startpoint
    for i in range(1, len(spikes)):
        if spikes[i] > 0:
            signal[i] = signal[i - 1] + threshold
        elif spikes[i] < 0:
            signal[i] = signal[i - 1] - threshold
        else:
            signal[i] = signal[i - 1]
    return signal


def ben_spike(spikes, fir, shift):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    #   Sengupta et al. (2017)
    #   Schrauwen et al. (2003)
    signal = np.convolve(spikes, fir)
    signal = signal + shift * np.ones(len(signal))
    signal = signal[0 : (len(signal) - len(fir) + 1)]
    return signal


def grf_spike(spikes, min_input, max_input):
    shape = spikes.shape
    signal = np.zeros(shape[0])
    for i in range(shape[0]):
        signal[i] = min_input + (2 * (np.argmax(spikes[i, :]) + 1) - 3) / 2 * (
            max_input - min_input
        ) / (shape[1] - 2)
    return signal


def one_hot_place_spike(spikes, min_input, max_input):
    shape = spikes.shape
    signal = np.zeros(shape[0])
    for i in range(shape[0]):
        signal[i] = min_input + (2 * (np.argmax(spikes[i, :]) + 1) - 3) / 2 * (
            max_input - min_input
        ) / (shape[1] - 2)
    return signal


def grf_spike_with_internal_timesteps(spikes, n_timesteps, min_input, max_input):
    shape = spikes.shape
    spikes = spikes.reshape((int(shape[0] / n_timesteps), n_timesteps, shape[1]))
    signal = np.zeros(len(spikes))
    mu = np.zeros(shape[1])

    for i in range(shape[1]):
        mu[i] = min_input + (2 * (i + 1) - 3) / 2 * (max_input - min_input) / (
            shape[1] - 2
        )

    for i in range(len(spikes)):
        spike_times = np.zeros(shape[1])
        for j in range(n_timesteps):
            for spike_idx in spikes[i, j, :].nonzero():
                spike_times[spike_idx] = n_timesteps - j

        weight_center = np.sum(mu * spike_times) / np.sum(spike_times)
        signal[i] = weight_center

    return signal



import logging
import multiprocessing as mp
import os
import sys
from glob import glob

import click
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from tqdm import tqdm

from miv.core import Spikestamps
from miv.signal.filter import ButterBandpass, FilterCollection, MedianFilter
from miv.signal.spike import ThresholdCutoff


def preprocessing(data):
    signal, timestamps, sampling_rate = data
    signal_filter = ButterBandpass(400, 1500, order=4)
    spike_detection = ThresholdCutoff(cutoff=5)
    filtered_signal = signal_filter(signal, sampling_rate)
    spikestamp = spike_detection(
        filtered_signal,
        timestamps,
        sampling_rate,
        return_neotype=False,
        progress_bar=False,
    )
    return spikestamp


@click.command()
@click.option("--path", "-p", help="Path to the recorded directory.", multiple=True)
@click.option(
    "--tools",
    "-t",
    type=click.Choice(["OpenEphys", "Intan"]),
    default="OpenEphys",
    help="Select the recording format",
)
@click.option(
    "--nproc",
    "-N",
    default=None,
    help="Number of processing units. If not provided, use maximum.",
)
@click.option(
    "--num_fragments",
    "-n",
    default=1,
    help="Number of fragments for data processing. Recommend to split data into 1 minute recording segments. For Intan, this number is pre-determined by the number of saved files.",
)
@click.option(
    "--use-mpi",
    type=bool,
    default=False,
    help="Set True if mpi is ready. Else, it will use multiprocessing. (mpi4py must be installed)",
)
@click.option("--chunksize", default=1, help="Number of chunks for multiprocessing")
def main(path, tools, nproc, num_fragments, use_mpi, chunksize):
    # Select tools
    if tools == "OpenEphys":
        # If OpenEphys is selected, process all recordings in the set.
        from miv.io import DataManager

        raise NotImplementedError
    elif tools == "Intan":
        from miv.io.intan import DataIntan

        for p in path:
            total_spikestamps = Spikestamps([])
            data = DataIntan(p)
            with mp.Pool(nproc) as pool:
                for spikestamp in pool.imap(
                    preprocessing, data.load(), chunksize=chunksize
                ):
                    total_spikestamps.extend(spikestamp)
            data.save_data(total_spikestamps, "spiketrain")

            logging.info(f"Pre-processing {p} done.")


if __name__ == "__main__":
    main()



from __future__ import annotations

__doc__ = """

Data Manager
############

.. autoclass:: DataManager
   :members:

Module (OpenEphys)
##################

.. Note::
    We expect the data structure to follow the default format
    exported from OpenEphys system:
    `format <https://open-ephys.atlassian.net/wiki/spaces/OEW/pages/491632/Data+format>`_.

.. currentmodule:: miv.io.data

.. autoclass:: Data
   :members:

"""

__all__ = ["Data", "DataManager"]


from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import logging
import os
import pickle
from collections.abc import MutableSequence
from glob import glob

import matplotlib.pyplot as plt
import numpy as np

from miv.io.binary import load_continuous_data, load_recording, load_ttl_event
from miv.io.protocol import DataProtocol
from miv.signal.filter.protocol import FilterProtocol
from miv.signal.spike.protocol import SpikeDetectionProtocol
from miv.statistics import firing_rates
from miv.typing import SignalType

if TYPE_CHECKING:
    import mpi4py


class Data:
    """Single data unit handler.

    Each data unit that contains single recording. This class provides useful tools,
    such as masking channel, export data, interface with other packages, etc.
    If you have multiple recordings you would like to handle at the same time, use
    `DataManager` instead.

    By default recording setup, the following directory structure is expected in ``data_path``::

        recording1                              # <- recording data_path
        ├── continuous
        │   └── Rhythm_FPGA-100.0
        │       ├── continuous.dat
        │       ├── synchronized_timestamps.npy
        │       └── timestamps.npy
        ├── events
        │   ├── Message_Center-904.0
        │   │   └── TEXT_group_1
        │   │       ├── channels.npy
        │   │       ├── text.npy
        │   │       └── timestamps.npy
        │   └── Rhythm_FPGA-100.0
        │       └── TTL_1
        │           ├── channel_states.npy
        │           ├── channels.npy
        │           ├── full_words.npy
        │           └── timestamps.npy
        ├── sync_messages.txt
        ├── structure.oebin
        └── analysis                            # <- post-processing result
            ├── spike_data.npz
            ├── plot
            ├── spike
            └── mea_overlay


        Parameters
        ----------
        data_path : str
    """

    def __init__(
        self,
        data_path: str,
    ):
        self.data_path: str = data_path
        self._analysis_path: str = os.path.join(data_path, "analysis")
        self.masking_channel_set: set[int] = set()

        os.makedirs(self._analysis_path, exist_ok=True)

    @property
    def analysis_path(self):
        """Default sub-directory path to save analysis results"""
        return self._analysis_path

    @analysis_path.setter
    def analysis_path(self, path):
        if not os.path.isdir(path):
            os.makedirs(path)
        self._analysis_path = path

    def save_figure(
        self,
        figure: plt.Figure,
        group: str,
        filename: str,
        savefig_kwargs: dict[Any, Any] | None = None,
    ):
        """Save figure in analysis sub-directory

        Parameters
        ----------
        figure : plt.Figure
        group : str
        filename : str
        savefig_kwargs : Optional[Dict[Any, Any]]
            Additional parameters to pass to `plt.savefig`.
        """
        if savefig_kwargs is None:
            savefig_kwargs = {}

        dirpath = os.path.join(self.analysis_path, group)
        os.makedirs(dirpath, exist_ok=True)

        filepath = os.path.join(dirpath, filename)
        plt.figure(figure)
        plt.savefig(filepath, **savefig_kwargs)

    def has_data(self, filename: str):
        """Check if the analysis data already saved

        Parameters
        ----------
        filename : str
            File name to check
        """
        filepath = os.path.join(self.analysis_path, filename + ".pkl")
        return os.path.exists(filepath)

    def save_data(
        self,
        data,
        filename: str,
        pkl_kwargs: dict[Any, any] | None = None,
    ):
        """Save analysis data into sub-directory

        Parameters
        ----------
        data :
        filename : str
        pkl_kwargs : Optional[Dict[Any, Any]]
            Additional parameters to pass to `plt.savefig`.
        """
        pkl_kwargs = pkl_kwargs or {}
        filepath = os.path.join(self.analysis_path, filename + ".pkl")
        with open(filepath, "wb") as output_file:
            pickle.dump(data, output_file, **pkl_kwargs)

    def load_data(
        self,
        filename: str,
        pkl_kwargs: dict[Any, any] | None = None,
    ):
        """Quick load pickled data (data saved using `save_data`)

        Parameters
        ----------
        filename : str
        pkl_kwargs : Optional[Dict[Any, Any]]
            Additional parameters to pass to `plt.savefig`.
        """
        pkl_kwargs = pkl_kwargs or {}
        filepath = os.path.join(self.analysis_path, filename + ".pkl")
        with open(filepath, "rb") as output_file:
            data = pickle.load(output_file, **pkl_kwargs)
        return data

    def load(
        self,
        num_fragments: int = 1,
        start_at_zero: bool = False,
        progress_bar=False,
        mpi_comm: mpi4py.MPI.COMM_WORLD | None = None,
    ):
        """
        Iterator to load data fragmentally.

        Parameters
        ----------
        num_fragments : int
            Instead of loading entire data at once, split the data into `num_fragment`
            number of subdata to process separately. (default=1)
        start_at_zero : bool
            If set to True, time first timestamps will be shifted to zero. To achieve synchronized
            timestamps with other recordings/events, set this to False.
        progress_bar : bool
            Visible progress bar
        mpi_comm: Optional[MPI.COMM_WORLD]
            (Experimental Feature) If the load is executed in MPI environment, user can pass
            MPI.COMM_WORLD to split the load.
            For example, if num_fragments=100, rank=2, and size=10, this function will only iterate
            the fragment 20-29 out of 0-100.

        Examples
        --------
            >>> data = Data(data_path)
            >>> for data.load(num_fragments=10) as (signal, timestamps, sampling_rate):
            ...     ...

        Returns
        -------
        signal : SignalType, neo.core.AnalogSignal
            The length of the first axis `signal.shape[0]` correspond to the length of the
            signal, while second axis `signal.shape[1]` correspond to the number of channels.
        timestamps : TimestampsType, numpy array
        sampling_rate : float

        Raises
        ------
        FileNotFoundError
            If some key files are missing.
        """
        # TODO: Not sure this is safe implementation
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")

        start_index = None
        end_index = None
        if mpi_comm is not None:
            rank = mpi_comm.Get_rank()
            size = mpi_comm.Get_size()
            tasks = np.array_split(np.arange(num_fragments), size)[rank]
            start_index = tasks.min()
            end_index = tasks.max() + 1
        yield from load_recording(
            self.data_path,
            self.masking_channel_set,
            start_at_zero=start_at_zero,
            num_fragments=num_fragments,
            start_index=start_index,
            end_index=end_index,
            progress_bar=progress_bar,
        )

    def load_ttl_event(self):
        """
        Load TTL event data if data contains. Detail implementation is :func:`here <miv.io.binary.load_ttl_event>`.
        """
        return load_ttl_event(self.data_path)

    def set_channel_mask(self, channel_id: Iterable[int]):
        """
        Set the channel masking.

        Parameters
        ----------
        channel_id : Iterable[int], list
            List of channel id that will be ignored.

        Notes
        -----
        If the index exceed the number of channels, it will be ignored.

        Examples
        --------
        >>> data = Data(data_path)
        >>> data.set_channel_mask(range(12,23))

        """
        self.masking_channel_set.update(channel_id)

    def clear_channel_mask(self):
        """
        Clears all present channel masks.
        """

        self.masking_channel_set = set()

    def _auto_channel_mask_with_correlation_matrix(
        self,
        spontaneous_binned: dict[str, Any],
        filter: FilterProtocol,
        detector: SpikeDetectionProtocol,
        offset: float = 0,
        bins_per_second: float = 100,
    ):
        """
        Automatically apply mask.

        Parameters
        ----------
        spontaneous_binned : Union[Iterable[Iterable[int]], int]
            [0]: 2D matrix with each column being the binned number of spikes from each channel.
            [1]: number of bins from spontaneous recording binned matrix
            [2]: array of indices of empty channels
        filter : FilterProtocol
            Filter that is applied to the signal before masking.
        detector : SpikeDetectionProtocol
            Spike detector that extracts spikes from the signals.
        offset : float, optional
            The trimmed time in seconds at the front of the signal (default = 0).
        bins_per_second : float, default=100
            Optional parameter for binning spikes with respect to time.
            The spikes are binned for comparison between the spontaneous recording and
            the other experiments. This value should be adjusted based on the firing rate.
            A high value reduces type I error; a low value reduces type II error.
            As long as this value is within a reasonable range, it should negligibly affect
            the result (see jupyter notebook demo).
        """

        exp_binned = self._get_binned_matrix(filter, detector, offset, bins_per_second)
        num_channels = np.shape(exp_binned["matrix"])[1]

        # if experiment is longer than spontaneous recording, it gets trunkated
        if exp_binned["num_bins"] > spontaneous_binned["num_bins"]:
            spontaneous_matrix = spontaneous_binned["matrix"].copy()
            exp_binned["matrix"] = exp_binned["matrix"][
                : spontaneous_binned["num_bins"] + 1
            ]

        # if spontaneous is longer than experiment recording
        elif exp_binned["num_bins"] < spontaneous_binned["num_bins"]:
            spontaneous_matrix = spontaneous_binned["matrix"].copy()
            spontaneous_matrix = spontaneous_matrix[: exp_binned["num_bins"] + 1]

        # they're the same size
        else:
            spontaneous_matrix = spontaneous_binned["matrix"].copy()

        exp_binned_channel_rows = np.transpose(exp_binned["matrix"])
        spontaneous_binned_channel_rows = np.transpose(spontaneous_matrix)

        dot_products = []
        for chan in range(num_channels):
            try:
                dot_products.append(
                    np.dot(
                        spontaneous_binned_channel_rows[chan],
                        exp_binned_channel_rows[chan],
                    )
                )
            except Exception:
                raise Exception(
                    "Number of channels does not match between this experiment and referenced spontaneous recording."
                )

        mean = np.mean(dot_products)
        threshold = mean + np.std(dot_products)

        mask_list = []
        for chan in range(num_channels):
            if dot_products[chan] > threshold:
                mask_list.append(chan)
        self.set_channel_mask(np.concatenate((mask_list, exp_binned["empty_channels"])))

    def _get_binned_matrix(
        self,
        filter: FilterProtocol,
        detector: SpikeDetectionProtocol,
        offset: float = 0,
        bins_per_second: float = 100,
    ) -> dict[str, Any]:
        """
        Performs spike detection and return a binned 2D matrix with columns being the
        binned number of spikes from each channel.

        Parameters
        ----------
        filter : FilterProtocol
            Filter that is applied to the signal before masking.
        detector : SpikeDetectionProtocol
            Spike detector that extracts spikes from the signals.
        offset : float, optional
            The time in seconds to be trimmed in front (default = 0).
        bins_per_second : float, default=100
            Optional parameter for binning spikes with respect to time.
            The spikes are binned for comparison between the spontaneous recording and
            the other experiments. This value should be adjusted based on the firing rate.
            A high value reduces type I error; a low value reduces type II error.
            As long as this value is within a reasonable range, it should negligibly affect
            the result (see jupyter notebook demo).

        Returns
        -------
        matrix :
            2D list with columns as channels.
        num_bins : int
            The number of bins.
        empty_channels : List[int]
            List of indices of empty channels
        """

        result = []
        for sig, times, samp in self.load(num_fragments=1):
            start_time = times[0] + offset
            starting_index = int(offset * samp)

            trimmed_signal = sig[starting_index:]
            trimmed_times = times[starting_index:]

            filtered_sig = filter(trimmed_signal, samp)
            spiketrains = detector(filtered_sig, trimmed_times, samp)

            bins_array = np.arange(
                start=start_time, stop=trimmed_times[-1], step=1 / bins_per_second
            )
            num_bins = len(bins_array)
            num_channels = len(spiketrains)
            empty_channels = []

            for chan in range(num_channels):
                if len(spiketrains[chan]) == 0:
                    empty_channels.append(chan)

                spike_counts = np.zeros(shape=num_bins + 1, dtype=int)
                digitized_indices = np.digitize(spiketrains[chan], bins_array)
                for bin_index in digitized_indices:
                    spike_counts[bin_index] += 1
                result.append(spike_counts)

        return {
            "matrix": np.transpose(result),
            "num_bins": num_bins,
            "empty_channels": empty_channels,
        }

    def save(self, tag: str, format: str):  # pragma: no cover
        # TODO
        assert tag == "continuous", "You cannot alter raw data, change the data tag"
        # save_path = os.path.join(self.data_path, tag)

        if format == "dat":
            ...
        elif format == "npz":
            ...
        elif format == "neo":
            ...
        else:
            raise NotImplementedError(
                "Format type " + format + " is not implemented.\n",
                "Please choose  one of the supported formats: dat, npz, neo",
            )

    def check_path_validity(self):
        """
        Check if necessary files exist in the directory.

        - Check `continious.dat` exists. (only one)
        - Check `structure.oebin` exists.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.

        """

        continuous_dat_paths = glob(
            os.path.join(self.data_path, "**", "continuous.dat"), recursive=True
        )
        if len(continuous_dat_paths) != 1:
            logging.warning(
                f"One and only one continuous.dat file can exist in the data path. Found: {continuous_dat_paths}"
            )
            return False
        if not os.path.exists(os.path.join(self.data_path, "structure.oebin")):
            logging.warning("Missing structure.oebin in the data path.")
            return False
        return True


class DataManager(MutableSequence):
    """
    Data collection manager.

    By default recording setup, the directory is named after the date and time
    of the recording. The structure of ``data_collection_path`` typically look
    like below::

        2022-03-10_16-19-09         <- data_collection_path
        └── Record Node 104
            └── experiment1
                └── recording1      <- data_path (Data module)
            ├── experiment2
            ├── experiment3
            ├── experiment4
            ├── spontaneous
            ├── settings.xml
            ├── settings_2.xml
            └── settings_3.xml

        Parameters
        ----------
        data_collection_path : str
            Path for data collection.

    """

    def __init__(self, data_collection_path: str):
        self.data_collection_path = data_collection_path
        self.data_list: list[DataProtocol] = []

        # From the path get data paths and create data objects
        self._load_data_paths()

    @property
    def data_path_list(self) -> Iterable[str]:
        return [data.data_path for data in self.data_list]

    # Queries
    def query_path_name(self, query_path) -> Iterable[DataProtocol]:
        return list(filter(lambda d: query_path in d.data_path, self.data_list))

    # DataManager Representation
    # TODO: Display data structure
    def tree(self):  # pragma: no cover
        """
        Pretty-print available recordings in DataManager in tree format.

        Examples
        --------
        >>> data_collection = DataManager("2022-05-15_13-51-36")
        >>> data_collection.tree()
        2022-05-15_14-51-36
            0: <miv.io.data.Data object at 0x7f8960660cd0>
               └── Record Node 103/experiment3_std2_pt_ESC/recording1
            1: <miv.io.data.Data object at 0x7f89671c8400>
               └── Record Node 103/experiment2_std1_pt_ESC/recording1
            2: <miv.io.data.Data object at 0x7f896199e7c0>
               └── Record Node 103/experiment1_cont_ESC/recording1

        """
        # TODO: Either use logging or other str stream
        if not self.data_list:
            print(
                "Data list is empty. Check if data_collection_path exists and correct"
            )
            return
        print(self.data_collection_path)
        for idx, data in enumerate(self.data_list):
            print(" " * 4 + f"{idx}: {data}")
            print(
                " " * 4
                + "   └── "
                + data.data_path[len(self.data_collection_path) + 1 :]
            )

    def _load_data_paths(self):
        """
        Create data objects from the data three.
        """
        # From the path get the data path list
        data_path_list = self._get_experiment_paths()

        # Create data object
        self.data_list = []
        invalid_count = 0
        for path in data_path_list:
            data = Data(path)
            if data.check_path_validity():
                self.data_list.append(data)
            else:
                invalid_count += 1
        logging.info(
            f"Total {len(data_path_list)} recording found. There are {invalid_count} invalid paths."
        )

    def _get_experiment_paths(self, sort: bool = True) -> Iterable[str]:
        """
        Get experiment paths.

        Parameters
        ----------
        sort : bool
            Sort the output data paths

        Returns
        -------
        data_path_list : list
        """
        # Use self.data_collection_path
        path_list = []
        for path in glob(
            os.path.join(self.data_collection_path, "*", "experiment*", "recording*")
        ):
            if (
                ("Record Node" in path)
                and ("experiment" in path)
                and os.path.isdir(path)
            ):
                path_list.append(path)
        if sort:
            path_list = sorted(path_list)
        return path_list

    def save(self, tag: str, format: str):  # pragma: no cover
        raise NotImplementedError  # TODO
        for data in self.data_list:
            data.save(tag, format)

    def apply_filter(self, filter: FilterProtocol):  # pragma: no cover
        raise NotImplementedError  # TODO
        for data in self.data_list:
            data.load()
            data = filter(data, sampling_rate=0)
            data.save(tag="filter", format="npz")
            data.unload()

    # MutableSequence abstract methods
    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx):
        return self.data_list[idx]

    def __delitem__(self, idx):
        del self.data_list[idx]

    def __setitem__(self, idx, data):
        if data.check_path_validity():
            self.data_list[idx] = data
        else:
            logging.warning("Invalid data cannot be loaded to the DataManager.")

    def insert(self, idx, data):
        if data.check_path_validity():
            self.data_list.insert(idx, data)
        else:
            logging.warning("Invalid data cannot be loaded to the DataManager.")

    def auto_channel_mask_with_firing_rate(
        self,
        filter: FilterProtocol,
        detector: SpikeDetectionProtocol,
        no_spike_threshold: float = 1,
    ):
        """
        Perform automatic channel masking.
        This method simply applies a Butterworth filter, extract spikes, and filter out
        the channels that contain either no spikes or too many spikes.

        Parameters
        ----------
        filter : FilterProtocol
            Filter that is applied to the signals before detecting spikes.
        detector : SpikeDetectionProtocol
            Spike detector that is used to extract spikes from the filtered signal.
        no_spike_threshold : float, default=1
            Spike rate threshold (spike per sec) for filtering channels with no spikes.
            (default = 1)

        """

        for data in self.data_list:
            for sig, times, samp in data.load(num_fragments=1):
                mask_list = []

                filtered_signal = filter(sig, samp)
                spiketrains = detector(filtered_signal, times, samp)
                spike_stats = firing_rates(spiketrains)

                for idx, channel_rate in enumerate(spike_stats["rates"]):
                    if int(channel_rate) <= no_spike_threshold:
                        mask_list.append(idx)

                data.set_channel_mask(mask_list)

    def auto_channel_mask_with_correlation_matrix(
        self,
        spontaneous_data: DataProtocol,
        filter: FilterProtocol,
        detector: SpikeDetectionProtocol,
        omit_experiments: Iterable[int] | None = None,
        spontaneous_offset: float = 0,
        exp_offsets: Iterable[float] | None = None,
        bins_per_second: float = 100,
    ):
        """
        This masking method uses a correlation matrix between a spontaneous recording and
        the experiment recordings to decide which channels to mask out.

        Notes
        -----
            Sample rate and number of channels for all recordings must be the same

        Parameters
        ----------
        spontaneous_data : Data
            Data from spontaneous recording that is used for comparison.
        filter : FilterProtocol
            Filter that is applied to the signals before detecting spikes.
        detector : SpikeDetectionProtocol
            Spike detector that is used to extract spikes from the filtered signal.
        omit_experiments: Optional[Iterable[int]]
            Integer array of experiment indices (0-based) to omit.
        spontaneous_offset: float, optional
            Postive time offset for the spontaneous experiment (default = 0).
            A negative value will be converted to 0.
        exp_offsets: Optional[Iterable[float]]
            Positive float array of time offsets for each experiment (default = 0).
            Negative values will be converted to 0.
        bins_per_second : float, default=100
            Optional parameter for binning spikes with respect to time.
            The spikes are binned for comparison between the spontaneous recording and
            the other experiments. This value should be adjusted based on the firing rate.
            A high value reduces type I error; a low value reduces type II error.
            As long as this value is within a reasonable range, it should negligibly affect
            the result (see jupyter notebook demo).
        """

        omit_experiments_list: list[float] = (
            list(omit_experiments) if omit_experiments else []
        )
        exp_offsets_list: list[float] = list(exp_offsets) if exp_offsets else []

        if spontaneous_offset < 0:
            spontaneous_offset = 0

        exp_offsets_length = sum(1 for e in exp_offsets_list)
        for i in range(exp_offsets_length):
            if exp_offsets_list[i] < 0:
                exp_offsets_list[i] = 0

        if exp_offsets_length < len(self.data_list):
            exp_offsets_list = np.concatenate(
                (
                    np.array(exp_offsets_list),
                    np.zeros(len(self.data_list) - exp_offsets_length),
                )
            )

        spontaneous_binned = spontaneous_data._get_binned_matrix(
            filter, detector, spontaneous_offset, bins_per_second
        )

        for (exp_index, data) in enumerate(self.data_list):
            if not (exp_index in omit_experiments_list):
                data._auto_channel_mask_with_correlation_matrix(
                    spontaneous_binned,
                    filter,
                    detector,
                    exp_offsets_list[exp_index],
                    bins_per_second,
                )



__doc__ = ""
__all__ = ["FilterProtocol"]

import typing
from typing import Callable, Optional, Protocol, Union

import numpy as np
import numpy.typing as npt

from miv.typing import SignalType


class FilterProtocol(Protocol):
    """Behavior definition of all filter operator."""

    tag: str = ""

    def __call__(self, array: SignalType, sampling_rate: float, **kwargs) -> SignalType:
        """User can apply the filter by directly calling.
        Parameters
        ----------
        array : SignalType {"numpy.ndarray", "neo.core.AnalogSignal"}
        samping_rate : float
        """
        ...

    def __repr__(self) -> str:
        """String representation for interactive debugging."""
        ...



# A script to convert Open Ephys data to MiV HDF5 format.

import logging
import os
import sys

import click
import numpy as np

from miv.io import Data, DataManager
from miv.io import file as miv_file


def config_logging(verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)


script_name = os.path.basename(__file__)


def seq_find(f, seq):
    """
    Determines if an element satisfying predicate is present in the
    given sequence, returns index of element or None.

    :param f: predicate function
    :param seq: sequence
    :return: index of element satisfying predicate, or None

    """
    i = 0
    for x in seq:
        if f(x):
            return i
        else:
            i = i + 1
    return None


@click.command()
@click.option("--folder-path", "-p", required=True, type=click.Path())
@click.option("--verbose", "-v", type=bool, default=False, is_flag=True)
def main(folder_path: str, verbose: bool):
    """Convert Open Ephys data to MiV HDF5 format"""

    config_logging(verbose)
    logger = logging.getLogger("miv_file")

    # Load dataset from OpenEphys recording
    data_manager = DataManager(folder_path)
    miv_data = miv_file.initialize()
    cont = miv_file.create_container(miv_data)

    # Get signal and rate(hz)
    #   signal        : np.array, shape(N, N_channels)
    #   timestamps    : np.array
    #   sampling_rate : float
    for recording in data_manager:
        data_path = recording.data_path
        with recording.load() as (signal, timestamps, sampling_rate):
            group_id = miv_file.create_group(miv_data, data_path, counter="nrecording")
            miv_file.create_dataset(
                miv_data,
                ["signal", "timestamps", "sampling_rate"],
                group=data_path,
                dtype=float,
            )
            cont[f"{group_id}/signal"] = signal
            cont[f"{group_id}/timestamps"] = timestamps
            cont[f"{group_id}/sampling_rate"] = [sampling_rate]
            miv_file.pack(miv_data, cont, logger=logger)

    miv_file.write(
        f"{folder_path}/MiV_data.h5",
        miv_data,
        comp_type="gzip",
        comp_opts=9,
        logger=logger,
    )


if __name__ == "__main__":
    main(
        args=sys.argv[
            (seq_find(lambda x: os.path.basename(x) == script_name, sys.argv) + 1) :
        ]
    )



__doc__ = """
Sample datasets for debugging/testing.

Any downloaded datasets will be saved in **datasets** directory.
If the dataset already exists, download will be skipped.
To remove any cache files, you can delete the **datasets** directory.
"""

from miv.datasets.optogenetic import load_data



__all__ = ["MedianFilter"]

from typing import Optional, Tuple, Union

import logging
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from miv.typing import SignalType


@dataclass(frozen=True)
class MedianFilter:
    """Median filter with threshold

    If the signal exceed the threshold, the value is replaced by median of neighboring
    values.

    The filter is designed to mitigate unusually high amplitude of spike within
    the signal.

    Parameters
    ----------
    threshold : float
        clipping threshold.
    k : Union[int, Tuple[int, int]]
        Number of neighboring values to take median. (default=20)
        It is possible to provide (k1, k2) tuple to separate right-neighbor bound
        and left-neighbor bound. If you are unsure which value to use, try with the
        value at least `k=0.002 * sampling_rate`.
    tag : str
        Tag for the collection of filter.
    """

    threshold: float
    k: Union[int, Tuple[int, int]] = 20
    tag: str = ""

    def __call__(
        self,
        signal: SignalType,
        *args,
        **kwargs,
    ) -> SignalType:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            pre-filtered signal

        Returns
        -------
        SignalType
            filtered signal

        """
        y = signal.copy()
        if isinstance(self.k, int):
            k = (self.k, self.k)
        else:
            k = self.k
        outlier_i, outlier_ch = np.where(np.abs(signal) > self.threshold)
        for i, ch in zip(outlier_i, outlier_ch):
            low_bound = np.max((0, i - k[0]))
            up_bound = np.min((i + k[1], signal.shape[0])) + 1
            y[low_bound:up_bound, ch] = np.median(signal[low_bound:up_bound, ch])
        return y

    def __post_init__(self):
        assert (
            self.threshold is None or self.threshold > 0.0
        ), f"Filter threshold {self.threshold} must be positive real number."
        if self.threshold is not None and self.threshold < 50.0:
            logging.warning(
                "Threshold is less than 50.0 uV, which could alter the spike signal"
            )



__doc__ = """
Lyon Ear Model

.. automodule:: LyonEarModel

"""
__all__ = ["LyonEarModel"]

from typing import Optional

import lyon.calc
import numpy as np


class LyonEarModel:  # TemporalEncoderProtocol
    """
    Lyon's Ear Model. Wrapper for lyon.LyonCalc module.

    Implementation details can be found `here <https://engineering.purdue.edu/~malcolm/interval/1998-010/>`_
    The above MATLAB/C implementation is ported to python `here <https://github.com/sciforce/lyon>`_

    .. automodule:: lyon.LyonCalc.lyon_passive_ear
    """

    def __init__(
        self,
        sampling_rate: float = 16000,
        decimation_factor: int = 1,
        ear_quality: int = 8,
        step_factor: Optional[float] = None,
        differ: bool = True,
        agc: bool = True,
        tau_factor: int = 3,
    ):
        """
        Auditory nerve response, based on Lyon's model.

        Parameters
        ----------
        sampling_rate : float
            sample_rate Waveform sample rate. (default=16000)
        decimation_factor : int
            decimation_factor How much to decimate model output. (default=1)
        ear_quality : int
            Smaller values mean broader filters. (default=8)
        step_factor : float
            Filter stepping factor. If not given, default value set to 25% overlap.
            (default=ear_quality / 32)
        differ : bool
            Differ Channel difference: improves model's freq response. (default=True)
        agc : bool
            Whether to use AGC for neural model adaptation. (default=True)
        tau_factor : int
            Reduces antialiasing in filter decimation. (default=3)
        """
        self.lyon = lyon.calc.LyonCalc()
        self.parameters = {
            "sample_rate": sampling_rate,
            "decimation_factor": decimation_factor,
            "ear_q": ear_quality,
            "step_factor": step_factor,
            "differ": differ,
            "agc": agc,
            "tau_factor": tau_factor,
        }

    def __call__(self, signal: np.ndarray) -> np.ndarray:
        """

        Parameters
        ----------
        signal : np.ndarray
            signal waveform: mono,

        Returns
        -------
        Output : np.ndarray
            array shape (N / decimation_factor, channels)
        """
        return self.lyon.lyon_passive_ear(signal, **self.parameters)



from miv.statistics.burst import *
from miv.statistics.info_theory import *
from miv.statistics.pairwise_causality import *
from miv.statistics.signal_statistics import *
from miv.statistics.spiketrain_statistics import *



__doc__ = """
Basic utilities for serial communication using PySerial package.
The tool is originally developped for experiment using `Stimjim <https://bitbucket.org/natecermak/stimjim/src/master/>`_
The purpose is to output spiketrain as pulse signal.

author: <Seung Hyun Kim skim0119@gmail.com>
"""
__all__ = ["ArduinoSerial", "list_serial_ports"]

import os
import sys
import time

import serial


def list_serial_ports():  # pragma: no cover
    """list serial communication ports available"""
    from serial.tools.list_ports import main

    main()


class ArduinoSerial:
    """
    Stimjim compatible
      - Baudrate: 112500
    """

    def __init__(self, port: str, baudrate: int = 112500):
        self._data_started = False
        self._data_buf = ""
        self._message_complete = False
        self.baudrate = baudrate
        self.port = port
        self.serial_port = None

    def connect(self):
        self.serial_port = self._setup_serial(self.baudrate, self.port)

    def _setup_serial(
        self, baudrate: int, serial_port_name: str, verbose: bool = False
    ):
        """Setup serial connection.

        Parameters
        ----------
        baudrate : int
            serial bits communication rate (bit per sec)
        serial_port_name : str
            Serial port name. Typically start with "COM". To scan available ports,
            run `list_serial_ports`.
        verbose : bool
            If set to true, print out debugging messages (default=False).
        """
        serial_port = serial.Serial(
            port=serial_port_name, baudrate=baudrate, timeout=0, rtscts=True
        )
        if verbose:
            print(f"{serial_port_name=} {baudrate=}")
        self.wait()
        return serial_port

    @property
    def is_open(self):
        return self.serial_port.is_open

    def open(self):
        self.serial_port.open()

    def close(self):
        self.serial_port.close()

    def send(
        self,
        msg: str,
        start_character: str = "",
        eol_character: str = "\n",
        verbose: bool = False,
    ):
        # adds the start- and end-markers before sending
        full_msg = start_character + msg + eol_character
        self.serial_port.write(full_msg.encode("utf-8"))
        if verbose:
            print(f"Msg send: {full_msg}")

    def receive(self, start_character="", eol_character="\n"):
        """receive.

        Parameters
        ----------
        start_character :
            start_character
        eol_character :
            eol_character
        """
        if self.serial_port.in_waiting() > 0 and not self._message_complete:
            x = self.serial_port.read().decode("utf-8")  # decode needed for Python3

            if self._data_started:
                if x != eol_character:
                    self._data_buf = self._data_buf + x
                else:
                    self._data_started = False
                    self._message_complete = True
            elif x == start_character:
                self._data_buf = ""
                self._data_started = True

        if self._message_complete:
            self._message_complete = False
            return self._data_buf
        else:
            return "ready"

    def wait(self, verbose: bool = False):
        """
        Allows time for Arduino launch. It also ensures that any bytes left
        over from a previous message are discarded
        """
        if verbose:
            print("Waiting for Arduino to reset")

        msg = ""
        prev_msg = ""
        while msg.lower().find("ready") == -1:
            msg = self.receive()
            prev_msg = msg
        return prev_msg



__doc__ = """

Signal Filter
#############

.. currentmodule:: miv.signal.filter

A sequence of filter can be composed using the module `FilterCollection`.

.. autoclass:: FilterCollection
   :members: append, insert

.. autosummary::
   :nosignatures:
   :toctree: _toctree/FilterAPI

   FilterProtocol
   ButterBandpass
   MedianFilter

"""
__all__ = ["FilterCollection"]

from typing import TYPE_CHECKING, List

from collections.abc import MutableSequence

import numpy as np
import numpy.typing as npt

from miv.typing import SignalType

if TYPE_CHECKING:
    from miv.signal.filter import FilterProtocol


class FilterCollection(MutableSequence):
    """Create sequence of multiple filters.

    Each appended filters are expected to abide the FilterProtocol.

        Parameters
        ----------
        tag : str
            Tag for the collection of filter.

        See Also
        --------
        miv.signal.filter.ButterBandpass : butterworth filter
        miv.signal.filter.MedianFilter : median filter with threshold

        Examples
        --------
        >>> custom_filter = (
        ...     FilterCollection(tag="custom filter 1")
        ...       .append(ButterBandpass(lowcut=300, highcut=3000, order=3))
        ...       .append(MedianFilter(threshold=50, k=50))
        ... )
        >>> custom_filter
        Collection of filters(self.tag='custom filter 1', id(self)=140410408570736)
          0: ButterBandpass(lowcut=300, highcut=3000, order=3, tag='')
          1: MedianFilter(threshold=50, k=50, tag='')
        >>> filtered_signal = custom_filter(signal=signal, sampling_rate=30_000)

    """

    def __init__(self, tag: str = ""):
        self.filters: List[FilterProtocol] = []
        self.tag: str = tag

    def __call__(self, signal: SignalType, sampling_rate: float) -> SignalType:
        for filt in self.filters:
            signal = filt(signal, sampling_rate)
        return signal

    # MutableSequence abstract methods
    def __len__(self):
        return len(self.filters)

    def __getitem__(self, idx):
        return self.filters[idx]

    def __delitem__(self, idx):
        del self.filters[idx]

    def __setitem__(self, idx, system):
        self.filters[idx] = system

    def insert(self, idx, system):
        """

        Parameters
        ----------
        idx : int
        system : FilterProtocol

        Returns
        -------
        self : FilterCollection
            User can chain multiple insert/append to build sequence of filter.
            Check the example above.

        """

        self.filters.insert(idx, system)
        return self

    def append(self, system):
        """

        Parameters
        ----------
        system : FilterProtocol

        Returns
        -------
        self : FilterCollection
            User can chain multiple insert/append to build sequence of filter.
            Check the example above.

        """
        self.filters.append(system)
        return self

    def __repr__(self):
        s = f"Collection of filters({self.tag=}, {id(self)=})\n"
        for idx, filter in enumerate(self.filters):
            s += f"  {idx}: {filter}\n"
        return s



__all__ = ["pairwise_causality_plot", "spike_triggered_average_plot"]

import os

import matplotlib.pyplot as plt
import numpy as np
from viziphant.spike_train_correlation import plot_corrcoef

from miv.statistics import pairwise_causality
from miv.statistics.spiketrain_statistics import binned_spiketrain
from miv.typing import SignalType, SpikestampsType


def pairwise_causality_plot(signal: SignalType, start: int, end: int):
    """
    Plots pairwise Granger Causality

    Parameters
    ----------
    signal : SignalType
        Input signal
    start : int
        Starting point of the signal
    end : int
        End point of the signal

    Returns
    -------
    figure : matplotlib.pyplot.figure
        Contains subplots for directional causalities for X -> Y and Y -> X,
        instantaneous causality between X,Y, and total causality. X and Y
        represents electrodes
    axes : matplotlib.axes
        axes parameters for plot modification

    See Also
    --------
    miv.statistics.pairwise_causality

    """

    # Causality
    corrcoef_mat = pairwise_causality(signal, start, end)

    # Plotting
    fig, axes = plt.subplots(2, 2)
    plt.subplots_adjust(
        left=None, bottom=None, right=None, top=None, wspace=0.6, hspace=0.6
    )
    plot_corrcoef(corrcoef_mat[0], axes=axes[0, 0])
    plot_corrcoef(corrcoef_mat[1], axes=axes[0, 1])
    plot_corrcoef(corrcoef_mat[2], axes=axes[1, 0])
    plot_corrcoef(corrcoef_mat[3], axes=axes[1, 1])
    axes[0, 0].set_xlabel("Electrode")
    axes[0, 1].set_xlabel("Electrode")
    axes[1, 0].set_xlabel("Electrode")
    axes[1, 1].set_xlabel("Electrode")
    axes[0, 0].set_ylabel("Electrode")
    axes[1, 0].set_ylabel("Electrode")
    axes[0, 1].set_ylabel("Electrode")
    axes[1, 1].set_ylabel("Electrode")
    axes[0, 0].set_title("Directional causality X => Y")
    axes[0, 1].set_title("Directional causality Y => X")
    axes[1, 0].set_title("Instantaneous causality of X,Y")
    axes[1, 1].set_title("Total interdependence of X,Y")

    return fig, axes


def spike_triggered_average_plot(
    signal: SignalType,
    channel_x: int,
    spiketrains: SpikestampsType,
    channel_y: int,
    sampling_freq: float,
    window_length: int,
):
    """
    Plots the spike-triggered average of Local Field Potential (LFP) from channel X
    corresponding to spiketrain from channel Y. The spiketrain from channel Y
    can be replaced with stimulation signal to understand stimulus dependent
    LFP on channel X, but take care in providing stimualtion as SpikestampsType.

    Parameters
    ----------
    signal : SignalType
        LFP signal recorded from the electrodes
    channel_x : float
        Channel to consider for LFP data
    spiketrains : SpikestampsType
        Single spike-stamps
    channel_y : float
        Channel to consider for spiketrain data
    sampling_freq : float
        sampling frequency for LFP recordings
    window_length : int
        window length to consider before and after spike

    Returns
    -------
    figure : matplotlib.pyplot.Figure
        matplot figure plotting spike triggered average of channel X in the provided window
    axes : matplotlib.axes.Axes
        axes parameters for plot modification

    """

    # Spike Triggered Average
    dt = 1.0 / sampling_freq
    n = np.shape(signal[:, channel_x])[0] / sampling_freq
    assert (
        window_length < np.shape(signal[:, channel_x])[0] / 2
    ), "Window cannot be longer than signal length"
    spike = binned_spiketrain(spiketrains[channel_x], 0, n, dt)
    lfp = signal[:, channel_x]
    spike_times = np.where(spike == 1)[0]
    spike_len = np.shape(spike_times)[0]
    sta = np.zeros([2 * window_length + 1])

    for i in np.arange(spike_len):
        sta += lfp[spike_times[i] - window_length : spike_times[i] + window_length + 1]

    spike_triggered_average = sta / spike_len
    lags = np.arange(-window_length, window_length + 1) * dt * 1000

    # Plotting
    fig, ax = plt.subplots()
    plt.plot(lags, spike_triggered_average)
    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Voltage (\u03bcV)")

    return fig, ax



__all__ = ["plot_frequency_domain", "plot_spectral"]

import matplotlib.pyplot as plt
import numpy as np
from scipy import fftpack
from scipy.signal import coherence, csd, welch

from miv.typing import SignalType


def plot_frequency_domain(signal: SignalType) -> plt.Figure:
    """
    Plot DFT frequency domain

    Parameters
    ----------
    signal : SignalType
        Input signal

    Returns
    -------
    figure: plt.Figure

    """
    # FFT
    # TODO: Move out the computational part into 'miv/signal'
    fig = plt.figure()
    sig_fft = fftpack.fft(signal)
    plt.plot(np.abs(sig_fft) ** 2)
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("DFT frequency")
    return fig


def plot_spectral(
    signal: SignalType, X: int, Y: int, sampling_rate: float, Number_Segments: int
):
    """
    Plots power spectral densities for channels X and Y: cross power spctral densities
    and coherence between them. [1]_, [2]_

    Parameters
    ----------
    signal : SignalType
        Input signal
    X : int
        First Channel
    Y : int
        Second Channel
    sampling_rate : float
        Sampling frequency
    Number_Segments: int
        Number of segments to divide the entire signal

    Returns
    -------
    figure: matplotlib.pyplot.Figure
    axes: matplotlib.Axes

    .. [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.coherence.html
    .. [2] P. Welch, “The use of the fast Fourier transform for the estimation of power spectra:
       A method based on time averaging over short, modified periodograms”, IEEE Trans. Audio
       Electroacoust. vol. 15, pp. 70-73, 1967.

    """
    ## Welch PSD for an electrode's Signal
    ## Welch Coherence Estimation between signal X and Y

    # TODO: Move out the computational part into 'miv/signal'

    L = np.int32(len(signal[:, 0]) / Number_Segments)  # L = length of each segment
    fs = sampling_rate  # fs = Sampling Frequeny

    fx, Pxx_den = welch(signal[:, X], fs, nperseg=L)  # Welch PSD and frequency for X
    fy, Pyy_den = welch(signal[:, Y], fs, nperseg=L)  # Welch PSD and frequency  for Y
    fxy, Pxy = csd(  # Welch CSD and frequency for X and Y
        signal[:, X], signal[:, Y], fs, nperseg=L
    )
    fcxy, Cxy = coherence(  # Welch Coherence and frequency for X and Y
        signal[:, X], signal[:, Y], fs, nperseg=L
    )

    # Plotting
    fig, axes = plt.subplots(2, 2)
    plt.subplots_adjust(
        left=None, bottom=None, right=None, top=None, wspace=0.6, hspace=0.6
    )
    axes[0, 0].semilogy(fx, Pxx_den)
    axes[0, 1].semilogy(fy, Pyy_den)
    axes[1, 0].semilogy(fxy, Pxy)
    axes[1, 1].semilogy(fcxy, Cxy)

    axes[0, 0].set_ylabel("PSD [V**2/Hz]")
    axes[1, 0].set_ylabel("CSD [V**2/Hz]")
    axes[1, 0].set_xlabel("Frequency [Hz]")
    axes[1, 1].set_xlabel("Frequency [Hz]")
    axes[0, 1].set_ylabel("PSD [V**2/Hz]")
    axes[1, 1].set_ylabel("Coherence")
    axes[0, 0].set_xlabel("Frequency [Hz]")
    axes[0, 1].set_xlabel("Frequency [Hz]")
    axes[0, 0].set_title("PSD for X")
    axes[0, 1].set_title("PSD for Y")
    axes[1, 0].set_title("CPSD for X and Y")
    axes[1, 1].set_title("Coherence for X,Y")
    return fig, axes



__all__ = [
    "shannon_entropy",
    "block_entropy",
    "entropy_rate",
    "active_information",
    "mutual_information",
    "relative_entropy",
    "conditional_entropy",
    "transfer_entropy",
]


from typing import Any, Dict, Iterable, List, Optional, Union

import datetime

import elephant.statistics
import matplotlib.pyplot as plt
import neo
import numpy as np
import pyinform
import quantities as pq
import scipy
import scipy.signal

from miv.statistics.spiketrain_statistics import binned_spiketrain
from miv.typing import SpikestampsType


def shannon_entropy(
    spiketrains: SpikestampsType,
    channel: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the shannon entropy for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channel : float
        electrode/channel
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
        shannon_entropy: float
        Shannon entropy for the given channel

    """

    bin_spike = binned_spiketrain(spiketrains, channel, t_start, t_end, bin_size)
    spike_dist = pyinform.dist.Dist(bin_spike)
    shannon_entropy = pyinform.shannon.entropy(spike_dist)
    return shannon_entropy


def block_entropy(
    spiketrains: SpikestampsType,
    channel: float,
    his: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the block entropy for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channel : float
        electrode/channel
    his : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    block_entropy: float
        Block entropy for the given channel

    """

    bin_spike = binned_spiketrain(spiketrains, channel, t_start, t_end, bin_size)
    block_entropy = pyinform.blockentropy.block_entropy(bin_spike, his)
    return block_entropy


def entropy_rate(
    spiketrains: SpikestampsType,
    channel: float,
    his: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the entropy rate for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channel : float
        electrode/channel
    his : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    entropy rate: float
        entropy rate for the given channel

    """

    bin_spike = binned_spiketrain(spiketrains, channel, t_start, t_end, bin_size)
    entropy_rate = pyinform.entropyrate.entropy_rate(bin_spike, k=his)
    return entropy_rate


def active_information(
    spiketrains: SpikestampsType,
    channel: float,
    his: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the active information for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channel : float
        electrode/channel
    his : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    active_information: float
        Active information for the given channel

    """

    bin_spike = binned_spiketrain(spiketrains, channel, t_start, t_end, bin_size)
    active_information = pyinform.activeinfo.active_info(bin_spike, his)
    return active_information


def mutual_information(
    spiketrains: SpikestampsType,
    channelx: float,
    channely: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the mutual information for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    mutual_information: float
        Mutual information for the given pair of electrodes

    """

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    mutual_information = pyinform.mutualinfo.mutual_info(bin_spike_x, bin_spike_y)
    return mutual_information


def relative_entropy(
    spiketrains: SpikestampsType,
    channelx: float,
    channely: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the relative entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    relative_entropy: float
        Relative_entropy for the given pair of electrodes

    """

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    relative_entropy = pyinform.relativeentropy.relative_entropy(
        bin_spike_x, bin_spike_y
    )
    return relative_entropy


def conditional_entropy(
    spiketrains: SpikestampsType,
    channelx: float,
    channely: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the conditional entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    conditional_entropy: float
        conditional entropy for the given pair of electrodes

    """

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    conditional_entropy = pyinform.conditionalentropy.conditional_entropy(
        bin_spike_x, bin_spike_y
    )
    return conditional_entropy


def transfer_entropy(
    spiketrains: SpikestampsType,
    channelx: float,
    channely: float,
    his: float,
    t_start: float,
    t_end: float,
    bin_size: float,
):
    """
    Estimates the transfer entropy for the pair of electorde recordings (X & Y) using the binned spiketrains and history

    Parameters
    ----------
    spiketrains : SpikestampsType
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    his : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    transfer_entropy: float
        Transfer_entropy for the given pair of electrodes

    """

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    transfer_entropy = pyinform.transferentropy.transfer_entropy(
        bin_spike_x, bin_spike_y, k=his
    )
    return transfer_entropy



__doc__ = """

Module (Intan)
##################

.. autoclass:: DataIntan
   :members:

"""
__all__ = ["DataIntan"]

from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import logging
import os
import pickle
import xml.etree.ElementTree as ET
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

import miv.io.intan.rhs as rhs
from miv.io.data import Data, DataManager
from miv.typing import SignalType


class DataManagerIntan(DataManager):
    pass


class DataIntan(Data):
    """Single data unit handler, collected from Intan.

    Each data unit that contains single recording. This class provides useful tools,
    such as masking channel, export data, interface with other packages, etc.
    If you have multiple recordings you would like to handle at the same time, use
    `DataManager` instead.

    By default recording setup, the following directory structure is expected in ``data_path``::

        data_<date>_<time>              # <- recorded data path
        ├── settings.xml
        ├── data_<date>_<time1>.rhs      # <- first rhs file
        ├── data_<date>_<time2>.rhs      # <- second rhs file
        ├── data_<date>_<time3>.rhs      # <- third rhs file
        ⋮
        └── analysis

        Parameters
        ----------
        data_path : str
    """

    def load(self):
        """
        Iterator to load data fragmentally.
        This function loads each file separately.

        Examples
        --------
            >>> data = Data(data_path)
            >>> for data.load(10) as (signal, timestamps, sampling_rate):
            ...     ...

        Returns
        -------
        signal : SignalType, neo.core.AnalogSignal
            The length of the first axis `signal.shape[0]` correspond to the length of the
            signal, while second axis `signal.shape[1]` correspond to the number of channels.
        timestamps : TimestampsType, numpy array
        sampling_rate : float

        Raises
        ------
        FileNotFoundError
            If some key files are missing.
        """

        yield from self._generator_by_channel_name("amplifier_data")

    def get_stimulation(self, progress_bar=False):
        """
        Load stimulation recorded data.
        """
        signals, timestamps = [], []
        for signal, timestamp, sampling_rate in self._generator_by_channel_name(
            "stim_data", progress_bar
        ):
            signals.append(signal)
            timestamps.append(timestamp)

        return (
            np.concatenate(signals, axis=0),
            np.concatenate(timestamps),
            sampling_rate,
        )

    def _generator_by_channel_name(self, name: str, progress_bar: bool = False):
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")
        files = self.get_recording_files()
        files.sort()
        # Get sampling rate from setting file
        setting_path = os.path.join(self.data_path, "settings.xml")
        sampling_rate = int(ET.parse(setting_path).getroot().attrib["SampleRateHertz"])
        # Read each files
        for filename in tqdm(files, disable=not progress_bar):
            result, data_present = rhs.load_file(filename)
            assert data_present, f"Data does not present: {filename=}."
            assert not hasattr(result, name), f"No {name} in the file ({filename=})."

            # signal_group = result["amplifier_channels"]
            yield np.asarray(result[name]).T, np.asarray(result["t"]), sampling_rate

    def check_path_validity(self):
        """
        Check if necessary files exist in the directory.

        - Check `rhs` file exists.
        - Check `settings.xml` file exists.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.
        """

        continuous_dat_paths = self.get_recording_files()
        if len(continuous_dat_paths) == 0:
            logging.warning("At least one .rhs file must exist in the data path.")
            return False
        if not os.path.exists(os.path.join(self.data_path, "settings.xml")):
            logging.warning("Missing settings.xml in the data path.")
            return False
        return True

    def get_recording_files(self):
        return glob(os.path.join(self.data_path, "*.rhs"), recursive=True)

    # Disable
    def load_ttl_event(self):
        raise AttributeError("DataIntan does not have laod_ttl_event method")



__doc__ = """
Spikestamps
===========

.. autoclass:: Spikestamps
   :members:

"""

__all__ = ["Spikestamps"]

from typing import Optional

from collections import UserList

import numpy as np


class Spikestamps(UserList):
    """List of array of spike times

    Represents spikes emitted by the same unit in a period of times.
    """

    def __init__(self, iterable):
        super().__init__()
        self.data = iterable

    def __setitem__(self, index, item):
        self.data[index] = item

    def insert(self, index, item):
        if index > len(self.data) or index < 0:
            raise IndexError("Index out of range")
        self.data.insert(index, item)

    def append(self, item):
        self.data.append(item)

    def extend(self, other):
        """
        Extend spikestamps from another `Spikestamps` or list of arrays.

        If the given parameter is another `Spikestamps`, each arrays are concatenated.
        Remaining channels will be added as different channels.

        If the given parameter is list of another arrays, each arrays will be added as different channels.

        Examples
        --------

        >>> a = Spikestamps([[0,1,2],[0,3]])
        >>> b = Spikestamps([[3,5],[4],[0,1]])
        >>> a.extend(b)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1]]

        >>> c = [[1],[5],[0]]
        >>> a.extend(c)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1], [1], [5], [0]]

        """
        if isinstance(other, type(self)):
            length_diff = len(other) - len(self.data)
            if length_diff > 0:
                for _ in range(length_diff):
                    self.data.append([])
            for idx, array in enumerate(other):
                self.data[idx].extend(array)
        else:
            self.data.extend(item for item in other)

    def get_count(self):
        """Return list of spike-counts for each channel."""
        return [len(data) for data in self.data]

    def get_last_spikestamp(self):
        """Return timestamps of the last spike in this spikestamps"""
        return max([data[-1] for data in self.data if len(data) > 0])

    def get_first_spikestamp(self):
        """Return timestamps of the first spike in this spikestamps"""
        return min([data[0] for data in self.data if len(data) > 0])

    def get_view(self, tstart: float, tend: float):
        """Truncate array and only includes spikestamps between tstart and tend."""
        return Spikestamps(
            [
                np.array(sorted(list(filter(lambda x: tstart <= x <= tend, arr))))
                for arr in self.data
            ]
        )



__all__ = ["signal_to_noise", "spike_amplitude_to_background_noise"]

import numpy as np

from miv.typing import SignalType, SpikestampsType
from miv.visualization.waveform import extract_waveforms


def signal_to_noise(signal: SignalType, axis: int = 0, ddof: int = 0):
    """
    Compute signal-to-noise ratio of raw signal.

    Parameters
    ----------
    signal : SignalType
    axis : int
        Axis of interest. By default, signal axis is 0 (default=1)
    ddof : int
    """
    signal_np = np.asanyarray(signal)
    m = signal_np.mean(axis)
    sd = signal_np.std(axis=axis, ddof=ddof)
    return np.abs(np.where(sd == 0, 0, m / sd))


def spike_amplitude_to_background_noise(
    signal: SignalType, spikestamps: SpikestampsType, sampling_rate: float
):
    """
    Calculate the signal-to-noise ratio (SNR) for a given signal and spikestamps.

    The SNR for each channel is defined as the mean spike amplitude squared, divided by the background noise amplitude.

    The definition is given `here <https://en.wikipedia.org/wiki/Signal-to-noise_ratio>`_.

    Parameters
    ----------
    signal: SignalType,
        The signal as a 2-dimensional numpy array (length, num_channel)
    spikestamps: SpikestampsType,
        The sample index of all spikes as a 1-dim numpy array
    sampling_rate : float
        The sampling frequency in Hz

    Returns
    -------
    List[float]: The SNR for each channel.
    """
    assert signal.shape[1] == len(
        spikestamps
    ), f"The number of channel for given signal {signal.shape[1]} is not equal to the number of channels in spikestamps {len(spikestamps)}."

    num_channels = signal.shape[1]
    snr = np.empty(num_channels)
    for channel in range(num_channels):
        # Compute mean spike amplitude
        if len(spikestamps[channel]) == 0:
            snr[channel] = np.nan
            continue
        cutouts = extract_waveforms(signal, spikestamps, channel, sampling_rate)
        mean_spike_amplitude = np.mean([np.max(np.abs(cutout)) for cutout in cutouts])

        # Compute background noise
        background_noise_amplitude = signal[:, channel].std()

        snr[channel] = (mean_spike_amplitude / background_noise_amplitude) ** 2
    return snr



#! /usr/bin/env python

# This file is part of the SpikeCoding repository - MAVLab TU Delft
#
# MIT License
#
# Copyright (c) 2021 Julien Dupeyroux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# @author Julien Dupeyroux, Stein Stroobants

import numpy as np
from scipy.stats import norm


def temporal_contrast(data, factor):
    # Based on algorithm provided in:
    #   Sengupta et al. (2017)
    #   Petro et al. (2020)
    diff = np.zeros(len(data)-1)
    spikes = np.zeros(len(data))
    for i in range(len(data)-1):
        diff[i] = data[i+1] - data[i]
    threshold = np.mean(diff) + factor * np.std(diff)
    diff = np.insert(diff, 0, diff[1])
    for i in range(len(data)):
        if diff[i] > threshold:
            spikes[i] = 1
        elif diff[i] < -threshold:
            spikes[i] = -1
    return spikes, threshold


def step_forward(data, threshold):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    startpoint = data[0]
    spikes = np.zeros(len(data))
    base = startpoint
    for i in range(1,len(data)):
        if data[i] > base + threshold:
            spikes[i] = 1
            base = base + threshold
        elif data[i] < base - threshold:
            spikes[i] = -1
            base = base - threshold
    return spikes, startpoint


def moving_window(data, threshold, window):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    startpoint = data[0]
    spikes = np.zeros(len(data))
    base = np.mean(data[0:window+1])
    for i in range(window+1):
        if data[i] > base + threshold:
            spikes[i] = 1
        elif data[i] < base - threshold:
            spikes[i] = -1
    for i in range(window+2, len(data)):
        base = np.mean(data[(i-window-1):(i-1)])
        if data[i] > base + threshold:
            spikes[i] = 1
        elif data[i] < base - threshold:
            spikes[i] = -1
    return spikes, startpoint


def hough_spike(data, fir):
    # Based on algorithm provided in:
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift*np.ones(len(data))
    for i in range(len(data)):
        count = 0
        for j in range(len(fir)):
            if i+j < len(data):
                if data[i+j] >= fir[j]:
                    count = count + 1
        if count == len(fir):
            spikes[i] = 1
            for j in range(len(fir)):
                if i+j < len(data):
                    data[i+j] = data[i+j] - fir[j]
    return spikes, shift


def modified_hough_spike(data, fir, threshold):
    # Based on algorithm provided in:
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift*np.ones(len(data))
    for i in range(len(data)):
        error = 0
        for j in range(len(fir)):
            if i+j < len(data):
                if data[i+j] < fir[j]:
                    error = error + fir[j] - data[i+j]
        if error <= threshold:
            spikes[i] = 1
            for j in range(len(fir)):
                if i+j < len(data):
                    data[i+j] = data[i+j] - fir[j]
    return spikes, shift


def ben_spike(data, fir, threshold):
    # Based on algorithm provided in:
    #   Petro et al. (2020)
    #   Sengupta et al. (2017)
    #   Schrauwen et al. (2003)
    spikes = np.zeros(len(data))
    shift = min(data)
    data = data - shift*np.ones(len(data))
    for i in range(len(data)-len(fir)+1):
        err1 = 0
        err2 = 0
        for j in range(len(fir)):
            err1 = err1 + abs(data[i+j] - fir[j])
            err2 = err2 + abs(data[i+j-1])
        if err1 <= err2*threshold:
            spikes[i] = 1
            for j in range(len(fir)):
                if i+j+1 < len(data):
                    data[i+j+1] = data[i+j+1] - fir[j]
    return spikes, shift


def grf_spike(data, m, min_input, max_input):
    # Adapted from algorithm provided in:
    #   Bohté et al. (2002)
    # Modifications: definition of sigma, removal of beta constant,
    #                and modified WTA process

    if np.isscalar(data):
        data = [data]

    spikes = np.zeros((len(data),m))
    neuron_outputs = np.zeros(m)

    for j in range(len(data)):
        for i in range(m):
            mu = min_input + (2*(i + 1)-3)/2*(max_input - min_input)/(m-2)
            sigma = (max_input - min_input)/(m-2)
            neuron_outputs[i] = norm.pdf(data[j], mu, sigma)

        spikes[j,np.argmax(neuron_outputs)] = 1
    return spikes

def one_hot_place_spike(data, m, min_input, max_input):
    # Simple population coding algorithm adapted from Stagsted et al. (2020) that represents inputs by a location. 
    # An input is assigned to the neuron that is closest to its value. 
    # Only one neuron fires at every timestep

    if np.isscalar(data):
        data = [data]

    spikes = np.zeros((len(data),m))

    for j in range(len(data)):
        size_change = 1/2*(max_input - min_input)/(m-2) # to make sure it has the same lower/upper bounds as the Bohte paper
        idx = int(np.round(((data[j] - (min_input - size_change)) / ((max_input + size_change) - (min_input - size_change))) * (m - 1)))
        spikes[j, idx] = 1
    
    return spikes


def grf_spike_with_internal_timesteps(data, min_input, max_input, neurons=10, timesteps=10, beta=1.5):
    """Create a series of spikes based on Gaussian Receptive Fields
    Adapted from algorithm provided in:
        Bohté et al. (2002)
    
    Keyword arguments:
    data -- 
    neurons -- numbers of neurons (default 10)
    timesteps -- number of timesteps (default 10)
    min_input -- minimal value
    max_input -- maximum value
    beta -- tuning parameter that determines the width of the receptive fields
    """

    if np.isscalar(data):
        data = [data]
        
    spikes = np.zeros((len(data), timesteps, neurons))
    responses = np.zeros(neurons)

    # Calculation of mu and sigma of the Gaussian receptive fields
    mu = min_input + (2*(np.arange(neurons)+1)-3)/2*(max_input - min_input)/(neurons-2)
    sigma = 1/beta*(max_input - min_input)/(neurons-2)
    max_prob = norm.pdf(mu[0], mu[0], sigma)

    for j in range(len(data)):
        for i in range(neurons):
            responses[i] = norm.pdf(data[j], mu[i], sigma)
            size_change = max_prob / (2 * timesteps)
            new = int(np.round(((responses[i] + size_change) / (max_prob + 2 * size_change) * (timesteps + 1)) + 0.0001)) # 0.0001 for roundoff errors...
            spiking_time = timesteps - new
            if spiking_time < timesteps - 1:
                spikes[j, spiking_time, i] = 1
    spikes = spikes.reshape([len(data) * timesteps, neurons])
    return spikes
:



__doc__ = ""
__all__ = ["ButterBandpass"]

from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.signal as sps

from miv.typing import SignalType


@dataclass(frozen=True)
class ButterBandpass:
    """Classical bandpass filter using `scipy` butterworth filter

    Parameters
    ----------
    lowcut : float
        low-pass frequency
    highcut : float
        high-pass frequency
    order : int
        The order of the filter. (default=5)
    tag : str
        Tag for the collection of filter.
    btype : str
        Filter type: bandpass, lowpass, highpass, bandstop. (default="bandpass")
        If set to lowpass, the critical frequency is set to 'highcut'.
        If set to highpass, the critical frequency is set to 'lowcut'.
        If set to bandpass or bandstop, the critical frequency is '[lowcut, highcut]'.
    """

    lowcut: float
    highcut: float
    order: int = 5
    tag: str = ""
    btype: str = "bandpass"

    def __call__(
        self,
        signal: SignalType,
        sampling_rate: float,
        **kwargs,
    ) -> SignalType:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            signal
        sampling_rate : float
            sampling_rate

        Returns
        -------
        SignalType

        """
        b, a = self._butter_bandpass(sampling_rate)
        y = signal.copy()
        if len(signal.shape) == 1:
            y = sps.lfilter(b, a, signal)
        elif len(signal.shape) == 2:
            for ch in range(signal.shape[1]):
                y[:, ch] = sps.lfilter(b, a, signal[:, ch])
        else:
            raise ValueError(
                "This filter can be only applied to 1D (signal) or 2D array (signal, channel)"
            )
        return y

    def __post_init__(self):
        assert (
            self.lowcut < self.highcut
        ), f"{self.lowcut=} should be lower than {self.highcut=}."
        assert self.order > 0 and isinstance(
            self.order, int
        ), f"Filter {self.order} must be an nonnegative integer."
        assert (
            min(self.lowcut, self.highcut) > 0
        ), "Filter critical frequencies must be greater than 0"

    def _butter_bandpass(self, sampling_rate: float):
        nyq = 0.5 * sampling_rate
        low = self.lowcut / nyq
        high = self.highcut / nyq
        if self.btype in ["bandpass", "bandstop"]:
            critical_frequency = [low, high]
        elif self.btype == "highpass":
            critical_frequency = low
        elif self.btype == "lowpass":
            critical_frequency = high
        else:
            raise ValueError("Unknown btype: %s" % self.btype)
        b, a = sps.butter(self.order, critical_frequency, btype=self.btype)
        return b, a

    def plot_frequency_response(self, sampling_rate: float):
        """plot_frequency_response

        Parameters
        ----------
        sampling_rate : float

        Returns
        -------
        plt.Figure
        """
        b, a = self._butter_bandpass(sampling_rate)
        w, h = sps.freqs(b, a)
        fig = plt.figure()
        plt.semilogx(w, 20 * np.log10(abs(h)))
        plt.title(
            f"Butterworth filter (order{self.order}) frequency response [{self.lowcut},{self.highcut}]"
        )
        plt.xlabel("Frequency")
        plt.ylabel("Amplitude")
        return fig



class ThresholdBasedRepresentation:
    """
    Threshold Based Representation encoding [1]_

    .. [1] Delbruck, T., & Lichtsteiner, P. (2007, May). Fast sensory motor control based on event-based hybrid neuromorphic-procedural system. In 2007 IEEE International Symposium on Circuits and Systems (pp. 845-848). IEEE.

    """

    def __init__(self, data, threshold=0.5):
        """
        :param numpy.ndarray data:
            See :py:mod: nucube.utils.ReadCSV
        :param float threshold:
            Threshold to cut off spike detection.
        """

        super().__init__()
        self.data = data
        self.threshold = threshold

    def get_spikes(self):
        """Returns TBR spikes.

        :rtype: numpy.ndarray
        :return: spikes

        >>> spikes = TBR()
        >>> spikes.get_spikes()
            [[[0 0 0 ..., 0 0 0]
            [1 1 1 ..., 1 1 1]
            [0 0 1 ..., 1 0 1]
            ...,
            [0 0 0 ..., 1 0 0]
            [0 0 0 ..., 0 0 1]
            [0 0 0 ..., 0 0 0]]
            |
            [[0 0 0 ..., 0 0 0]
            [0 0 0 ..., 0 0 0]
            [0 1 0 ..., 0 1 1]
            ...,
            [0 0 0 ..., 0 0 0]
            [0 0 0 ..., 0 0 1]
            [1 0 1 ..., 0 0 1]]]

        """

        sample_feature_threshold = self._get_threshold()
        row_diff = self._get_training_element_difference()

        spikes = []
        for idx, sample in enumerate(row_diff):
            spikes.append(
                np.where(
                    np.greater(sample, sample_feature_threshold[idx]),
                    1,
                    np.where(np.less(sample, -sample_feature_threshold[idx]), -1, 0),
                )
            )

        # np.array([np.sign(sample_feature_threshold[i] - a) for i, a in enumerate(row_diff)])  # Possible answer

        return np.asarray(spikes, dtype=np.int8)

    def get_spike_time(self, offset=100):
        """
        Merges the samples and returns the time at which the spike happens. Spike time always startes with ``1`` and
        increases with :param: offset.

        :param int offset:
            Shifting the spike index to the given number, if the first index is ``0``, then it will always be ``1``
        :rtype: list
        :return: Spike time with an offset.
        """

        reshaped_raw_spikes = np.vstack(self.get_spikes())

        spike_times = []

        for _feature in reshaped_raw_spikes.T:
            _idx_of_ones = np.squeeze(np.add(np.where(_feature == 1), 1))
            if _idx_of_ones[0] == 1:
                spike_times.append(
                    np.insert(np.multiply(_idx_of_ones[1:], offset), 0, 1).tolist()
                )
            else:
                spike_times.append(np.multiply(_idx_of_ones, offset).tolist())

        return spike_times

    def _get_training_element_difference(self):
        """Returns row-by-row difference for each sample.

        :rtype: numpy.ndarray
        :return: row_diff
        """

        train_data = self.data

        row_diff = []
        for index in range(train_data.shape[0]):
            row_diff.append(np.diff(train_data[index], axis=0))

        return np.asarray(row_diff)

    def _get_mean_sd_sample(self):
        """Returns mean and standard deviation of absolute data.

        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: mean, sd
        """

        diff_data = self._get_training_element_difference()
        abs_data = np.abs(diff_data)
        abs_data_shape_range = range(abs_data.shape[0])

        mean = []
        sd = []
        for index in abs_data_shape_range:
            mean.append(np.mean(abs_data[index], axis=0))

        for index in abs_data_shape_range:
            sd.append(np.std(abs_data[index], axis=0))

        return np.asarray(mean), np.asarray(sd)

    def _get_threshold(self):
        """Returns threshold values for each feature of every sample.

        :rtype: ndarray
        :return: threshold
        """

        mean, sd = self._get_mean_sd_sample()

        return np.add(mean, np.multiply(sd, self.threshold))



__doc__ = """Sample TTL-input signal readout."""
__all__ = ["load_data"]

import gzip
import os

import numpy as np

from miv.datasets.utils import get_file
from miv.io import DataManager


def load_data():  # pragma: no cover
    """
    Loads the sample TTL data readout. `Direct Download <https://uofi.box.com/shared/static/9llg11ods9iejdt2omjwjosbsxb5ui10.zip>`_

    Total size: 17.4 kB (compressed)

    File hash: a4314442fd9eba4377934c5766971ba1e04f079b7e615b8fb033992323afeb3f

    Examples
    --------
        >>> from miv.datasets.ttl_events import load_data
        >>> experiments: miv.io.DataManager = load_data()
        datasets/ttl_recording/sample_event_recording
            0: <miv.io.data.Data object at 0x7fec71c99fa0>
               └── Record Node 101/experiment1/recording1


    Notes
    -----
    All experiment are 1 minute long, 30k Hz recording of optogenetic
    neuron cells over 64 channels MEA. Dataset includes 1 spontaneous
    recording and 4 stimulated recording.

    Spontaneous recording is the recording over 1 minute period without
    external stimulation. The purpose was to measure the baseline mean-
    firing rate.

    Stimulation was done by LED light. Over 1 minute (60 seconds) period,
    6 stimulation was done with 10 seconds of intervals. For each stimulation,
    LED light was shined over 1 seconds, followed by remaining 9 seconds
    of rest (without light).

    Containing experiments:

    * experiment0: TTL recording

    Returns
    -------
    dataset: miv.io.DataManager

    Examples
    --------
        >>> from miv import datasets
        >>> experiments: miv.io.DataManager = datasets.ttl_events.load_data()

    """

    subdir = "ttl_recording"
    base_url = "https://uofi.box.com/shared/static/w3cylplece450up6t6h53vuq93t98q2k.zip"
    file = "sample_event_recording.zip"
    file_hash = "a4314442fd9eba4377934c5766971ba1e04f079b7e615b8fb033992323afeb3f"

    path = get_file(
        file_url=base_url,
        directory=subdir,
        fname=file,
        file_hash=file_hash,
        archive_format="zip",
    )
    experiment = DataManager(path)
    experiment.tree()
    return experiment
